from expression import core,Function,WrappedType,ReplacementMap

class Evaluator(object):

    def __init__(self,evaluator,recursive = False,split_binary = True, split_binary_size = 2,S = None):

        if S == None:
            raise ValueError('missing argument S')

        self._evaluator = evaluator
        self._evaluator.settings.recursive = recursive
        self._evaluator.settings.split_binary = split_binary
        self._evaluator.settings.split_binary_size = split_binary_size

        self.S = S

    def __call__(self,expr,cache = None):
        if cache is not None:
            if not isinstance(cache,dict):
                raise ValueError('cache error: expected dict, got %s' % type(cache))
            c = cache.get(self)
            if not c:
                c = ReplacementMap()
                cache[self] = c
            return self.S(self._evaluator(self.S(expr),c._replacement_map))
        else:
            return self.S(self._evaluator(self.S(expr)))

class Rule(object):
    
    def __init__(self,search,replacement = None,evaluator=None,condition=None,S=None):
        if S == None:
            raise ValueError('missing argument S')
        
        if evaluator is not None:
            if condition is not None:
                self._rule = core.Rule(S(search),S(replacement),S(condition),S(True),lambda m:evaluator(ReplacementMap(m,S=S)))
            else:
                self._rule = core.Rule(S(search),S(replacement),lambda m:evaluator(ReplacementMap(m,S=S)))
        elif replacement is not None:
            if condition is not None:
                self._rule = core.Rule(S(search),S(replacement),S(condition),S(True))
            else:
                self._rule = core.Rule(S(search),S(replacement))
        else:
            self._rule = search    
        self.S = S
    
    @property
    def search(self):
        return self.S(self._rule.search)

    @property
    def has_evaluator(self):
        return self._rule.has_evaluator()

    @property
    def replacement(self):
        return self.S(self._rule.replacement)

    @property
    def condition(self):
        c = self._rule.get_condition()
        if c:
            return self.S(c)

    def __repr__(self):
        l = str(self.search) + ' -> ' + str(self.replacement)
        if self.has_evaluator:
            l += ' ...'
        return l
    
    def _repr_latex_(self):
        l = self.search._repr_latex_()[2:-2] + r' \rightarrow ' + self.replacement._repr_latex_()[2:-2]
        if self.has_evaluator:
            l += '\; \dots '
        return "$$%s$$" % l

WrappedRule = lambda S:WrappedType(Rule,S=S)

class MatchCondition(Function):
    
    def __init__(self,name,F,S):
        super(MatchCondition,self).__init__(core.MatchCondition(name,lambda e:F(S(e))),S=S)

WrappedMatchCondition = lambda S:WrappedType(MatchCondition,S=S)

class ReplaceEvaluator(Evaluator):

    def __init__(self,replacement_map = None,**kwargs):
        if replacement_map is not None:
            super(ReplaceEvaluator,self).__init__(core.ReplaceEvaluator(replacement_map),**kwargs)
        else:
            super(ReplaceEvaluator,self).__init__(core.ReplaceEvaluator(),**kwargs)

    def add_replacement(self,search,replace):
        self._evaluator.add_replacement(self.S(search),self.S(replace))

WrappedReplaceEvaluator = lambda S:WrappedType(ReplaceEvaluator,S=S)

class RewriteEvaluator(Evaluator):

    def __init__(self,**kwargs):
        super(RewriteEvaluator,self).__init__(core.RuleEvaluator(),**kwargs)

    def __len__(self):
        return len(self._evaluator)
    
    def __getitem__(self,idx):
        return Rule(self._evaluator.get_rule(idx),S=self.S)
    
    def __iter__(self):
        def generator():
            for i in range(len(self)):
                yield self[i]
        return generator()
    
    def add_rule(self,search,replace = None,evaluator=None,priority = 0,**kwargs):
        self._evaluator.add_rule(Rule(search,replace,evaluator,S=self.S,**kwargs)._rule,priority)
    
    def set_rule_callback(self,f):
        if f is not None:
            self._evaluator.set_apply_callback(lambda r,m:f(Rule(r,S=self.S),ReplacementMap(m,S=self.S)))
        else:
            self._evaluator.set_apply_callback(None)


WrappedRewriteEvaluator = lambda S:WrappedType(RewriteEvaluator,S=S)
    
class MultiEvaluator(Evaluator):
    
    def __init__(self,S = None,**kwargs):
        super(MultiEvaluator,self).__init__(core.MultiEvaluator(),S=S,**kwargs)
        self._inner_evaluators = []
    
    def add_evaluator(self,evaluator):
        self._inner_evaluators.append(evaluator)
        self._evaluator.add_evaluator(evaluator._evaluator)

    def set_rule_callback(self,callback):
        for e in self._inner_evaluators:
            if isinstance(e,(RewriteEvaluator,MultiEvaluator)):
                e.set_rule_callback(callback)

WrappedMultiEvaluator = lambda S:WrappedType(MultiEvaluator,S=S)


class StepEvaluator(Evaluator):

    def __init__(self,S = None,**kwargs):
        super(StepEvaluator,self).__init__(core.StepEvaluator(),S=S,**kwargs)
        self._inner_evaluators = []

    def add_evaluator(self,evaluator):
        self._inner_evaluators.append(evaluator)
        self._evaluator.add_evaluator(evaluator._evaluator)

    def set_rule_callback(self,callback):
        for e in self._inner_evaluators:
            if isinstance(e,(RewriteEvaluator,MultiEvaluator)):
                e.set_rule_callback(callback)

WrappedStepEvaluator = lambda S:WrappedType(StepEvaluator,S=S)


    
    
    




import expresso.pycas as pc
import rule_symbols as s

evaluator = pc.RewriteEvaluator(recursive=True, split_binary=True)

from .logic_evaluator import is_explicit_natural

def unary_rule(f):

    def evaluator(m):
        try:
            res = f(m[s.x].value)
        except:
            return False

        if res is None:
            return False
        else:
            m[s.z] = pc.S(res)

    return evaluator

def binary_rule(f):

    def evaluator(m):
        try:
            res = f(m[s.x].value,m[s.y].value)
        except:
            return False
        if res is None:
            return False
        else:
            m[s.z] = pc.S(res)

    return evaluator

def are_explicit_numbers(*args):
    return pc.And(*[is_explicit_natural(arg) for arg in args])

evaluator.add_rule(s.x + s.y, s.z, binary_rule(lambda x, y: x + y), condition=are_explicit_numbers(s.x, s.y))
evaluator.add_rule(-s.x-s.y, -s.z, binary_rule(lambda x, y: x + y), condition=are_explicit_numbers(s.x, s.y))
evaluator.add_rule(s.x - s.y, s.z, binary_rule(lambda x, y: x - y), condition=are_explicit_numbers(s.x, s.y))
evaluator.add_rule(s.x * s.y, s.z, binary_rule(lambda x, y: x * y), condition=are_explicit_numbers(s.x, s.y))
evaluator.add_rule(s.x ** s.a * s.y ** s.a, s.z ** s.a, binary_rule(lambda x, y: x * y), condition=are_explicit_numbers(s.x, s.y))

def exp_length(a,b):
    import math as m
    return m.log10(a)*b

evaluator.add_rule(s.x ** s.y, s.z, binary_rule(lambda x, y: x ** y if exp_length(x,y)<100 else None), condition=are_explicit_numbers(s.x, s.y))
evaluator.add_rule(s.x ** -s.y, s.z ** -1, binary_rule(lambda x, y: x ** y if exp_length(x,y)<100 else None), condition=are_explicit_numbers(s.x, s.y))

evaluator.add_rule(pc.mod(s.x,s.y), s.z, binary_rule(lambda x, y: x % y), condition=are_explicit_numbers(s.x, s.y))
evaluator.add_rule(pc.mod(s.x,-s.y), s.z, binary_rule(lambda x, y: x % (-y)), condition=are_explicit_numbers(s.x, s.y))
evaluator.add_rule(pc.mod(-s.x,s.y), s.z, binary_rule(lambda x, y: (-x) % (-y)), condition=are_explicit_numbers(s.x, s.y))
evaluator.add_rule(pc.equal(s.x,s.y), s.z, binary_rule(lambda x, y: x == y), condition=are_explicit_numbers(s.x, s.y))
evaluator.add_rule(pc.equal(s.x,-s.y), False, condition=are_explicit_numbers(s.x, s.y))

evaluator.add_rule(s.x < s.y, s.z, binary_rule(lambda x, y: x < y), condition=are_explicit_numbers(s.x, s.y))
evaluator.add_rule(s.x < -s.y, s.z, binary_rule(lambda x, y: x < -y), condition=are_explicit_numbers(s.x, s.y))
evaluator.add_rule(-s.x < s.y, s.z, binary_rule(lambda x, y: -x < y), condition=are_explicit_numbers(s.x, s.y))


#evaluator.add_rule(s.a**s.x*s.b**-s.x, (s.a*s.b**-1)**(s.x),condition=are_explicit_numbers(s.a, s.b))

def is_even(x):
    return pc.equal(pc.mod(x,2),0)

from type_evaluator import issubtype

def is_uneven(x):
    return pc.And(issubtype(x,pc.Types.Integer),pc.Not(is_even(x)))


evaluator.add_rule((-s.x)**s.y, s.x**s.y,condition=is_even(s.y))
evaluator.add_rule((-s.x)**s.y, -(s.x**s.y),condition=is_uneven(s.y))

evaluator.add_rule(s.x<s.y*s.z**-1,s.z*s.x<s.y ,condition=are_explicit_numbers(s.y,s.z))
evaluator.add_rule(s.x*s.z**-1<s.y,s.x<s.y*s.z ,condition=are_explicit_numbers(s.x,s.z))
evaluator.add_rule(s.x<-(s.y*s.z**-1),s.z*s.x<-s.y ,condition=are_explicit_numbers(s.y,s.z))
evaluator.add_rule(-(s.x*s.z**-1)<s.y,-s.x<s.y*s.z ,condition=are_explicit_numbers(s.x,s.z))

evaluator.add_rule(s.x<s.z**-1,s.z*s.x<1 ,condition=is_explicit_natural(s.z))
evaluator.add_rule(s.z**-1<s.y,1<s.y*s.z ,condition=is_explicit_natural(s.z))
evaluator.add_rule(s.x<-(s.z**-1),s.z*s.x<-1 ,condition=is_explicit_natural(s.z))
evaluator.add_rule(-(s.z**-1)<s.y,-1<s.y*s.z ,condition=is_explicit_natural(s.z))


evaluator.add_rule(pc.equal(s.x,s.y*s.z**-1),pc.equal(s.z*s.x,s.y),condition=are_explicit_numbers(s.y,s.z))
evaluator.add_rule(pc.equal(s.x,-(s.y*s.z**-1)),pc.equal(s.z*s.x,-s.y),condition=are_explicit_numbers(s.y,s.z))

evaluator.add_rule(pc.equal(s.x,-(s.y*s.z**-1)),pc.equal(s.z*s.x,-s.y),condition=are_explicit_numbers(s.y,s.z))


from .logic_evaluator import is_pure_numeric,is_numeric

evaluator.add_rule(s.a+s.b*s.c**-1,(s.a*s.c+s.b)*s.c**-1,condition=pc.And(is_pure_numeric(s.a),are_explicit_numbers(s.b,s.c)))
evaluator.add_rule(s.a-s.b*s.c**-1,(s.a*s.c-s.b)*s.c**-1,condition=pc.And(is_pure_numeric(s.a),are_explicit_numbers(s.b,s.c)))

evaluator.add_rule(s.a+s.c**-1,(s.a*s.c+1)*s.c**-1,condition=pc.And(is_pure_numeric(s.a),is_explicit_natural(s.c)))
evaluator.add_rule(s.a-s.c**-1,(s.a*s.c-1)*s.c**-1,condition=pc.And(is_pure_numeric(s.a),is_explicit_natural(s.c)))


evaluator.add_rule((s.a+s.b)**2,(s.a**2+2*s.a*s.b+s.b**2),condition=pc.And(is_numeric(s.a),is_numeric(s.b)))
evaluator.add_rule(s.a*(s.b+s.c),s.a*s.b+s.a*s.c,condition=pc.And(is_pure_numeric(s.a),is_numeric(s.b),is_numeric(s.c)))

evaluator.add_rule(pc.I**s.n,(-1)**(s.n/2),condition=is_even(s.n))
evaluator.add_rule(pc.I**s.n,pc.I*(-1)**(s.n/2-0.5),condition=is_uneven(s.n))

evaluator.add_rule(pc.sign(-s.x),-1,condition=is_explicit_natural(s.x))
evaluator.add_rule(pc.sign(s.x),1,condition=is_explicit_natural(s.x))

import fractions

def evaluate_fraction(m):
    vx = m[s.x].value
    vy = m[s.y].value
    res = fractions.Fraction(vx,vy)
    if (res.numerator,res.denominator) == (vx,vy):
        return False
    m[s.a] = res.numerator
    m[s.b] = res.denominator

evaluator.add_rule(s.x*s.y**-1,s.a*s.b**-1,evaluate_fraction,condition=are_explicit_numbers(s.x, s.y))

evaluator.add_rule(s.x**s.c*s.y**-s.c,s.a**s.c*s.b**-s.c,evaluate_fraction,condition=are_explicit_numbers(s.x, s.y))



from .canonical_form import canonical_form
from .logic_evaluator import logic_evaluator

numeric_evaluator = pc.MultiEvaluator(recursive=True, split_binary=True)
numeric_evaluator.add_evaluator(canonical_form)
numeric_evaluator.add_evaluator(logic_evaluator)
numeric_evaluator.add_evaluator(evaluator)









import ctypes
from ..codeprinter import CodePrinter
from .. import expression as e
from .. import functions as f
import expresso.visitor as visitor
from mpmath import mp


class c_complex(ctypes.Structure):
    _fields_ = [('real',ctypes.c_double),('imag',ctypes.c_double)]
    def __str__(self):
        return str(self.real) + '+' + str(self.imag) + 'j'
    def __repr__(self):
        return '(' + str(self.real) + ',' + str(self.imag) + ')'
    def __complex__(self):
        return complex(self.real,self.imag)
    def is_complex(self):
        return True
    @staticmethod
    def np_type():
        import numpy
        return numpy.complex128

class CCodePrinter(CodePrinter):

    def __init__(self):
        super(CCodePrinter,self).__init__()
        self.includes = {'cmath','complex','thread','future','vector'}
        self.namespaces = {'std'}

        self.typenames = {
          f.Types.Boolean:'bool',
          f.Types.Natural:'unsigned',
          f.Types.Integer:'int',
          f.Types.Rational:'double',
          f.Types.Real:'double',
          f.Types.Complex:'c_complex',
          None:'c_complex'
        }

        self.ctype_map = {
            'bool':ctypes.c_bool,
            'unsigned':ctypes.c_uint,
            'int':ctypes.c_int,
            'double':ctypes.c_double,
            'c_complex':c_complex
        }

        self.type_converters = {}
        self.need_conversion = {}
        self.preamble = set()
        self.globals = set()

        c_complex_type = '''
struct c_complex{ double real; double imag; };
inline std::complex<double> to_cpp_complex(c_complex z){ return std::complex<double>(z.real,z.imag); }
inline c_complex to_c_complex(std::complex<double> z){ return c_complex{z.real(),z.imag()}; }
        '''

        self.preamble.add(c_complex_type)
        self.type_converters['c_complex'] = (lambda x:'to_cpp_complex(%s)' % x,lambda x:'to_c_complex(%s)' % x)

        complex_operators = '''
inline complex<double> operator{0}(complex<double> lhs, const double & rhs){{
    return lhs {0} complex<double>(rhs);
}}
inline complex<double> operator{0}(const double & lhs,complex<double> rhs){{
    return complex<double>(lhs) {0} rhs;
}}
        '''

        self.preamble.update(set([complex_operators.format(op) for op in ['+', '-', '*', '/']]))

        parallel_for = '''
  inline unsigned hardware_thread_count(){ return std::thread::hardware_concurrency(); }

template<typename C1,typename C2,typename F> void parallel_for(C1 start,C2 end,F f,uintptr_t thread_count = hardware_thread_count()){
    if(end-start < thread_count) thread_count = end-start;

    std::vector<std::future<void>> handles(thread_count);
    C2 block_size = (end - start)/thread_count;
    for(uintptr_t i=0;i<thread_count-1;++i){
      handles[i] = std::async(std::launch::async,[=](){
        C2 begin = start+block_size*i, end = start+block_size*(i+1);
        for(C2 j=begin;j<end;++j){ f(j); }
      });
    }
    handles[thread_count-1] = std::async([&](){
      C2 begin = start+block_size*(thread_count-1);
      for(C2 j=begin;j<end;++j)f(j);
     });
    for(auto & handle:handles) handle.wait();
}
        '''

        self.preamble.add(parallel_for)

        ndarray = '''
template<size_t _size,size_t _stride, size_t... sizes> struct ndarray_index_calculator {
  using rest = ndarray_index_calculator<sizes...>;
  static size_t size(){ return _size; }
  template <typename ... Args> static bool is_valid(size_t idx,Args ... args){ if(!rest::is_valid(args...)) return false; return idx < size(); }
  static size_t stride(){ return _stride; }
  template <typename ... Args> static size_t get_index(size_t idx,Args ... args){ return stride() * idx + rest::get_index(args...); }
};

template<size_t _size,size_t _stride> struct ndarray_index_calculator <_size,_stride> {
  static size_t size(){ return _size; }
  static bool is_valid(size_t idx){ return idx < size(); }
  static size_t stride(){ return _stride; }
  static size_t get_index(size_t idx){ return idx; }
};

template <class T,size_t ... size_stride> struct mapped_ndarray{
  T * data;
  T default_value;
  using index_calculator = ndarray_index_calculator<size_stride...>;
  mapped_ndarray(T * d,const T &_default_value = 0):data(d),default_value(_default_value){ }
  template <typename ... Args> T & operator()(Args ... indices){
    if(!index_calculator::is_valid(indices...)){ return default_value; }
    return data[index_calculator::get_index(indices...)];
  }
};
        '''

        self.preamble.add(ndarray)

        self.preamble.add('''template <typename T> int sign(T val) { return (T(0) <= val) - (val < T(0)); }''')


    def needs_brackets_in(self,expr,parent):
        if expr.is_atomic:
            return False
        return expr.function.is_operator

    @visitor.on('expr',parent = CodePrinter)
    def visit(self,expr):
        raise ValueError('cannot print expression %s' % expr)

    @visitor.function(f.CustomFunction)
    def visit(self,expr):
        f = expr.args[0].value
        if hasattr(f,'ccode'):
            self.preamble.add(f.ccode)
        else:
            raise ValueError('cannot compile custom function %s' % expr)
        return "%s(%s)" % (f.name,','.join([self(arg) for arg in expr.args[1:]]))

    @visitor.function(f.exponentiation)
    def visit(self,expr):
        return 'pow(%s,%s)' % (self(expr.args[0]),self(expr.args[1]))

    @visitor.atomic(f.I)
    def visit(self,expr):
        return "std::complex<double>(0,1)"

    @visitor.atomic(f.pi)
    def visit(self,expr):
        return "M_PI"

    @visitor.atomic(f.e)
    def visit(self,expr):
        return "M_E"

    @visitor.function(f.Xor)
    def visit(self,expr):
        return self.print_binary_operator(expr,symbol='^')

    @visitor.function(f.Not)
    def visit(self,expr):
        return "!(%s)" % self(expr.args[0])

    @visitor.function(f.equal)
    def visit(self,expr):
        return self.print_binary_operator(expr,'==')

    @visitor.function(f.fraction)
    def visit(self,expr):
        return "1./(%s)" % self(expr.args[0])

    @visitor.function(f.mod)
    def visit(self,expr):
        return "fmod(%s,%s)" % (self(expr.args[0]),self(expr.args[1]))

    @visitor.function(f.InnerPiecewise)
    def visit(self,expr):
        parts = ['(%s)?(%s):' % (self(arg.args[1]),self(arg.args[0])) for arg in expr.args]
        return '(%s%s)' % (''.join(parts),self(e.S(0)))

    @visitor.symbol
    def visit(self,expr):
        converter = self.need_conversion.get(expr)
        if converter:
            if isinstance(converter,tuple):
                return converter[0](expr)
            else:
                return converter(expr)
        return expr.name

    @visitor.atomic(e.S(True))
    def visit(self,expr):
        return 'true'

    @visitor.atomic(e.S(False))
    def visit(self,expr):
        return 'false'

    def print_includes(self):
        return '\n'.join(['#include <%s>' % name for name in self.includes])

    def print_namespaces(self):
        return '\n'.join(['using namespace %s;' % namespace for namespace in self.namespaces])

    def print_auxiliary_code(self):
        return '%s\n%s' % ('\n'.join(self.preamble),'\n'.join(self.globals))

    def print_file(self,*function_definitions):

        function_code = [self.generate_function(f) for f in function_definitions]
        function_code += [self.generate_vector_function(f,use_previous_definition=True) for f in function_definitions]

        return "\n\n".join([self.print_includes(),
                            self.print_namespaces(),
                            self.print_auxiliary_code()] + function_code )

    def print_typename(self,expr):
        return self.typenames.get(expr,self.typenames[None])

    def print_vector_typename(self,expr):
        return "%s*" % self.typenames.get(expr,self.typenames[None])

    def get_ctype(self,typename):
        if typename[-1] == '*':
            return ctypes.POINTER(self.get_ctype(typename[:-1]))
        return self.ctype_map[typename]

    @visitor.function(f.unfoldable)
    def visit(self,expr):
        return self.visit(expr.args[0])

    @visitor.function(f.ArrayAccess)
    def visit(self,expr):
        arr = expr.args[0].value
        pointer = arr.ctypes.data
        type = f.type_converters.numpy_c_typenames[arr.dtype.name]
        size = ','.join(["%s,%s" % (size,stride/arr.itemsize) for size,stride in zip(arr.shape,arr.strides)])
        name = expr.args[0].name
        self.globals.add('mapped_ndarray<%s,%s> %s((%s*)%s);' % (type,size,name,type,pointer))
        return "%s(%s)" % (name,','.join([self(arg) for arg in reversed(expr.args[1:])]))

    @visitor.obj(mp.mpf)
    def visit(self,expr):
        return repr(float(expr.value))

    @visitor.obj(mp.mpc)
    def visit(self,expr):
        v = expr.value
        return "complex<double>(%s,%s)" % (repr(float(v.real)),repr(float(v.imag)))

    def optimize_function(self,expr):
        from expresso.pycas.evaluators.optimizers import optimize_for_compilation
        return optimize_for_compilation(expr)

    def get_body_code(self,definition):
        if definition.return_type == None:
            return_type = self.print_typename(f.Type(definition.expr).evaluate())
        else:
            return_type = self.print_typename(definition.return_type)
        f_code = self(self.optimize_function(definition.expr))
        if return_type in self.type_converters and isinstance(self.type_converters[return_type],tuple):
            f_code = self.type_converters[return_type][1](f_code)
        return f_code

    def generate_function(self,definition):

        if definition.return_type == None:
            return_type = self.print_typename(f.Type(definition.expr).evaluate())
        else:
            return_type = self.print_typename(definition.return_type)

        args = definition.args

        if definition.arg_types == None:
            argument_types = [self.print_typename(f.Type(arg).evaluate()) for arg in args]
        else:
            argument_types = [self.print_typename(f.Type(arg).evaluate()) for arg in definition.arg_types]

        self.need_conversion = {arg:self.type_converters[t]
                                for arg,t in zip(args,argument_types)
                                if t in self.type_converters}

        f_code = self.get_body_code(definition)

        formatted = (return_type, definition.name,
                    ','.join(['%s %s' % (type,arg.name) for arg,type in zip(args,argument_types)]),
                    f_code)

        definition.c_return_type = self.get_ctype(return_type)
        definition.c_arg_types = [self.get_ctype(arg_type) for arg_type in argument_types]

        return 'extern "C"{\n%s %s(%s){\n\treturn %s;\n}\n}' % formatted

    def vectorized_name(self,name):
        return "__%s_vector" % name

    def generate_vector_function(self,definition,use_previous_definition = False):

        if definition.return_type == None:
            return_type = self.print_vector_typename(f.Type(definition.expr).evaluate())
        else:
            return_type = self.print_vector_typename(definition.return_type)

        args = definition.args

        if definition.arg_types == None:
            argument_types = [self.print_vector_typename(f.Type(arg).evaluate()) for arg in args]
        else:
            argument_types = [self.print_vector_typename(f.Type(arg).evaluate()) for arg in definition.arg_types]

        self.need_conversion.update({arg:lambda a:'%s[__i]' % a for arg in args})

        argument_types = ['unsigned',return_type] + argument_types

        if not use_previous_definition :
            f_code = self.get_body_code(definition)
        else:
            f_code = '%s(%s)' % (definition.name,','.join(self(arg) for arg in definition.args))

        if definition.parallel:
            f_code = 'parallel_for(0,__size,[&](unsigned __i){ __res[__i] = %s; }); ' % f_code
        else:
            f_code = 'for(unsigned __i = 0; __i<__size;++__i) __res[__i] = %s;' % f_code

        rargument_types = [argument_types[0]] + ['%s __restrict__ ' % t for t in argument_types[1:]]
        formatted_args = ','.join(['%s %s' % vardef for vardef in
                                   zip(rargument_types,['__size','__res'] + list(args))])

        formatted = (self.vectorized_name(definition.name), formatted_args, f_code)

        definition.c_vectorized_arg_types = [self.get_ctype(arg_type) for arg_type in argument_types]

        return 'extern "C"{\nvoid %s(%s){\n\t%s\n}\n}' % formatted


def ccompile(*function_definitions,**kwargs):
    import tempfile
    import shutil
    import ctypes
    import numpy as np
    from subprocess import Popen, PIPE
    from os import environ

    ccode_printer = CCodePrinter()
    code  = ccode_printer.print_file(*function_definitions)

    output_directory = tempfile.mkdtemp()

    object_file = output_directory+'/'+'pycas_compiled_expression.o'

    flags = kwargs.pop('flags',[])
    p = Popen([environ.get('CXX','g++'),'-o',object_file] + flags + ['-c','-xc++','-std=c++11','-ffast-math','-O3','-fPIC','-'],stdin=PIPE, stdout=PIPE, stderr=PIPE)
    p.stdin.write(code)
    p.stdin.close()

    return_code = p.wait()
    if(return_code!=0):
        raise RuntimeError("Cannot compile expression: " + p.stderr.read() )

    print_output = kwargs.pop('print_output',False)
    print_warnings = print_output or kwargs.pop('print_warnings',False)

    if print_warnings:
        print p.stderr.read()

    if print_output:
        print p.stdout.read()

    shared_library = output_directory+'/'+'pycas_compiled_expression.so'
    p = Popen(['g++','-shared','-o',shared_library,object_file],stdin=PIPE, stdout=PIPE, stderr=PIPE)
    p.stdin.close()

    return_code = p.wait()
    if(return_code!=0):
        raise RuntimeError("Cannot convert to shared library: " + p.stderr.read())

    if print_warnings:
        print p.stderr.read()

    if print_output:
        print p.stdout.read()

    lib = ctypes.cdll.LoadLibrary(shared_library)
    shutil.rmtree(output_directory)

    compiled_functions = {}

    class CompiledFunction(object):
        def __init__(self,cf,cf_vector):
            self.cf = cf
            self.cf_vector = cf_vector

        def __call__(self,*args,**kwargs):
            if(len(args) == 0):
                return self.cf()
            if any([isinstance(arg,(list,tuple)) for arg in args]):
                argtypes = self.cf_vector.argtypes
                args = [np.array(arg,dtype=t) for t,arg in zip(argtypes[2:],args)]
            if any([isinstance(arg,np.ndarray) for arg in args]):
                argtypes = self.cf_vector.argtypes

                shape = None
                for arg in args:
                    if isinstance(arg,np.ndarray):
                        if shape == None:
                            shape = arg.shape
                        else:
                            if arg.shape != shape:
                                raise AttributeError('c function got arguments with different shapes')

                args = [arg if isinstance(arg,np.ndarray) else arg * np.ones(shape) for arg in args]
                args = [np.ascontiguousarray(arg,dtype=t._type_) for t,arg in zip(argtypes[2:],args)]

                if argtypes[1]._type_ == c_complex:
                    restype = c_complex.np_type()
                else:
                    restype = argtypes[1]._type_

                res = kwargs.get('res')
                if res is None:
                    res = np.zeros(args[0].shape,dtype=restype)
                else:
                    assert res.dtype == restype
                    assert res.shape == shape
                    assert res.flags['C_CONTIGUOUS']

                call_args  = [res.size,res.ctypes.data_as(argtypes[1])]
                call_args += [arg.ctypes.data_as(t) for t,arg in zip(argtypes[2:],args)]
                self.cf_vector(*call_args)
                return res
            return self.cf(*args)

        def address(self):
            return ctypes.cast(self.cf, ctypes.c_void_p).value

    class CompiledLibrary(object):
        def __init__(self,lib,code):
            self.lib = lib
            self.code = code

    res = CompiledLibrary(lib,code)

    for definition in function_definitions:
        f = getattr(lib,definition.name)
        f.argtypes = definition.c_arg_types
        f.restype  = definition.c_return_type

        f_vector = getattr(lib, ccode_printer.vectorized_name(definition.name))
        f_vector.argtypes = definition.c_vectorized_arg_types
        f_vector.restype  = None

        setattr(res,definition.name,CompiledFunction(f,f_vector))

    return res






import expresso

Number = long

def Symbol(name,type=None,positive = False,latex = None,repr = None):
    s = Expression(expresso.create_symbol(name))
    if type != None:
        from functions import Type
        global_context.add_definition(Type(s),type)
    if positive == True:
        from .functions import sign
        global_context.add_definition(s>=0,True)
        global_context.add_definition(sign(s),1)
    if latex is not None:
        latex_rep = latex
        from printer import latex,add_target
        @add_target(latex,s)
        def print_latex(printer,expr):
            return latex_rep
    if repr is not None:
        from printer import printer,add_target
        @add_target(printer,s)
        def evaluate(printer,expr):
            return repr
    return s

def expression_converter(expr):
    if isinstance(expr,expresso.core.Expression):
        return Expression(expr)
    if isinstance(expr,Expression):
        return expr
    if isinstance(expr,bool):
        if expr == True:
            return Expression(expresso.create_object(expr))
        if expr == False:
            return Expression(expresso.create_object(expr))
    if isinstance(expr,(int)):
        expr = Number(expr)
    if isinstance(expr, Number):
        if expr >= 0:
            return Expression(expresso.create_object(expr))
        else:
            expr = abs(expr)
            return negative(expresso.create_object(expr))
    if isinstance(expr,float):
        import fractions
        f = fractions.Fraction(repr(expr))
        if f.denominator == 1:
            return expression_converter(f.numerator)
        if f.numerator == 1:
            return fraction(f.denominator)
        else:
            return f.numerator * fraction(f.denominator)
    if isinstance(expr,complex):
        if expr.real == 0:
            if expr.imag == 0:
                return Zero
            return I * S(float(expr.imag))
        if expr.imag == 0:
            return S(float(expr.real))
        return S(float(expr.real)) + I * S(float(expr.imag))
    if isinstance(expr,tuple):
        return Tuple(*expr)
    raise ValueError('Unsupported expression type: %s (%s)' % (type(expr),expr))

def S(value):
    if isinstance(value,str):
        return Symbol(value)
    return expression_converter(value)
    
def Wildcard(name):
    return S(expresso.core.create_wildcard_symbol(name))

def WildcardFunction(name):
    return expresso.expression.Function(expresso.core.WildcardFunction(name),S=expression_converter)

def symbols(string,**kwargs):
    return [Symbol(s.strip(),**kwargs) for s in string.split(',')]

def wildcard_symbols(string):
    string = string.replace(" ", "")
    return [Wildcard(s.strip()) for s in string.split(',')]

printer = expresso.printer.Printer(expression_converter)
latex = expresso.printer.LatexPrinter(expression_converter)

class Expression(expresso.WrappedExpression(expression_converter)):
    
    def __add__(self, other):
        return addition(self,other)

    def __radd__(self, other):
        return addition(other,self)

    def __neg__(self):
        return negative(self)

    def __pos__(self):
        return self
    
    def __sub__(self, other):
        return addition(self, negative(other))

    def __rsub__(self, other):
        return addition(other, negative(self))

    def __mul__(self, other):
        return multiplication(self,other)
    
    def __rmul__(self, other):
        return multiplication(other,self)

    def __div__(self, other):
        return multiplication(self, fraction(other))
    
    def __rdiv__(self, other):
        other = self.S(other)
        if other == One:
            return fraction(self)
        return multiplication(other, fraction(self))
    
    def __pow__(self,other):
        return exponentiation(self,other)

    def __rpow__(self,other):
        return exponentiation(other,self)

    def __mod__(self,other):
        return mod(self,other)

    def __rmod__(self,other):
        return mod(other,self)

    def __lt__(self, other):
        return Less(self,other)

    def __le__(self, other):
        return LessEqual(self,other)
    
    def __gt__(self, other):
        return Greater(self,other)

    def __ge__(self, other):
        return GreaterEqual(self,other)

    def __or__(self, other):
        return Or(self,other)

    def __xor__(self, other):
        return Xor(self,other)

    def __and__(self, other):
        return And(self,other)

    def __abs__(self):
        return Abs(self)
        
    def __nonzero__(self):
        raise ValueError('Cannot determine truth value of Expression. Perhaps you are using a python operator incorrectly?')
    
    def _repr_latex_(self):
         return "$$%s$$" % latex(self)

    def __repr__(self):
         return printer(self)

    def __iter__(self):
        from .functions import Tuple
        if self.function == Tuple:
            return self.args.__iter__()
        raise AttributeError('%s object has no attribute __iter__' % type(self))

    def evaluate(self,context = None,**kwargs):
        if context == None:
            context = global_context
        from evaluators import evaluate
        return evaluate(self,context = context,**kwargs)
    
    def subs(self,*args,**kwargs):
        #do_evaluate = kwargs.pop('evaluate',True)
        res = self.replace(*args)
        return res

    def N(self,prec = 16,**kwargs):
        from compilers import N
        return N(self,mp_dps=prec,**kwargs)

    def approximate(self,prec = 16,**kwargs):
        from expresso.pycas.evaluators.optimizers import optimize_for_compilation
        return optimize_for_compilation(self,prec=prec,**kwargs)

    def __float__(self):
        v = self.evaluate().N()
        try:
            return float(v)
        except:
            raise RuntimeError('expression %s is not convertable to float' % self)

    def __complex__(self):
        return complex(self.evaluate().N())

    def __int__(self):
        from compilers import lambdify
        v = lambdify(self.evaluate())()
        try:
            return int(v)
        except:
            raise RuntimeError('expression %s is not convertable to int' % self)

    def __long__(self):
        from compilers import lambdify
        v = lambdify(self.evaluate())()
        try:
            return long(v)
        except:
            raise RuntimeError('expression %s is not convertable to long' % self)

locals().update(expresso.WrappedExpressionTypes(Expression).__dict__)

class Context(ReplaceEvaluator):

    def add_definition(self,search,replacement):
        self.add_replacement(search,replacement)

global_context = Context()

One = S(1)
Zero = S(0)
NaN = S( expresso.create_object(float('nan'),'undefined value') )
I = S( expresso.create_object(1j,'imaginary unit') )

addition = BinaryOperator("+",expresso.associative,expresso.commutative,-11)
negative = UnaryOperator("-",expresso.prefix,-12)
multiplication = BinaryOperator("*",expresso.associative,expresso.commutative,-13)
fraction = UnaryOperator("1/",expresso.prefix,-14)
exponentiation = BinaryOperator("**",-15)

addition_group = Group(addition,negative,Zero)
multiplication_group = Group(multiplication,fraction,One)
real_field = Field(addition_group,multiplication_group)
complex_field = Field(addition_group,multiplication_group)

Or = BinaryOperator("|",expresso.associative,expresso.commutative,-3)
And = BinaryOperator("&",expresso.associative,expresso.commutative,-3)
Xor = BinaryOperator(" XOR ",expresso.associative,expresso.commutative,-3)

Not = UnaryOperator("~",expresso.prefix,-7)
mod = Function('mod',argc = 2)

equal = BinaryOperator("=",expresso.associative,expresso.commutative,-6)
unequal = BinaryOperator("!=",expresso.associative,expresso.commutative,-6);

In = BinaryOperator(" in ",-6)
NotIn = BinaryOperator(" not in ",-6)

Less = BinaryOperator("<",expresso.associative,expresso.non_commutative,-6)
LessEqual = BinaryOperator("<=",expresso.associative,expresso.non_commutative,-6)
Greater = BinaryOperator(">",expresso.associative,expresso.non_commutative,-6)
GreaterEqual = BinaryOperator(">=",expresso.associative,expresso.non_commutative,-6)

Abs = Function('abs',argc = 1)
Tuple = Function('tuple')






import expresso.pycas as pc
import rule_symbols as s

logic_evaluator = pc.RewriteEvaluator(recursive=True,split_binary=True)
evaluator = logic_evaluator

is_atomic = pc.Function('is atomic')
is_symbol = pc.Function('is symbol')
is_explicit_natural = pc.Function('is explicit natural')
is_pure_numeric = pc.Function('is pure numeric')
is_numeric = pc.Function('is numeric')
is_mpmath= pc.Function('is mpmath')

is_function = pc.Function('is function')
contains_atomic = pc.Function('contains atomic')

def is_explicit_natural_evaluator(m):
    for expr in m:
        if not isinstance(expr[1].value, pc.Number):
            m[s.y] = False
            return
    m[s.y] = True

logic_evaluator.add_rule(is_explicit_natural(s.x),s.y,is_explicit_natural_evaluator)

def is_function_type(expr,function):
    return is_function(expr,pc.expresso.create_object(function))

def comp_function_evaluator(m):
    f =  m[s.x].function
    if f is None or (isinstance(m[s.y].value,pc.Function) and f != m[s.y].value):
        m[s.z] = False
        return
    m[s.z] = True

logic_evaluator.add_rule(is_function(s.x,s.y),s.z,comp_function_evaluator)

def is_atomic_evaluator(m):
    if not m[s.x].is_atomic:
        m[s.y] = False
        return
    m[s.y] = True

logic_evaluator.add_rule(is_atomic(s.x),s.y,is_atomic_evaluator)
logic_evaluator.add_rule(is_function(s.x),pc.Not(is_atomic(s.x)))

def is_symbol_evaluator(m):
    if not m[s.x].is_symbol:
        m[s.y] = False
        return
    m[s.y] = True

logic_evaluator.add_rule(is_symbol(s.x),s.y,is_symbol_evaluator)


def is_pure_numeric_evaluator(m):
    m[s.y] = pc.And(*[is_pure_numeric(arg) for arg in m[s.x].args])

logic_evaluator.add_rule(is_pure_numeric(s.x),True,condition=is_explicit_natural(s.x))
logic_evaluator.add_rule(is_pure_numeric(s.x),False,condition=is_symbol(s.x))
logic_evaluator.add_rule(is_pure_numeric(s.x),s.y,is_pure_numeric_evaluator,condition=is_function(s.x))

from mpmath import mp

def is_mpmath_evaluator(m):
    v = m[s.x].value
    if isinstance(v,(mp.mpf,mp.mpc)):
        m[s.y] = True
    else:
        m[s.y] = False

logic_evaluator.add_rule(is_mpmath(s.x),s.y,is_mpmath_evaluator,condition=is_atomic(s.x))

def is_numeric_evaluator(m):
    m[s.y] = pc.And(*[is_numeric(arg) for arg in m[s.x].args])

logic_evaluator.add_rule(is_numeric(s.x),True,condition=is_explicit_natural(s.x))
logic_evaluator.add_rule(is_numeric(pc.pi),True)
logic_evaluator.add_rule(is_numeric(pc.e),True)
logic_evaluator.add_rule(is_numeric(pc.I),True)
logic_evaluator.add_rule(is_numeric(s.x),False,condition=is_symbol(s.x))
logic_evaluator.add_rule(is_numeric(s.x),s.y,is_numeric_evaluator,condition=is_function(s.x))


def contains_evaluator(m):
    search = m[s.y]
    m[s.z] = pc.Or(*[contains_atomic(arg, search) for arg in m[s.x].args])

logic_evaluator.add_rule(contains_atomic(s.x, s.x), True)
logic_evaluator.add_rule(contains_atomic(s.x, s.y), False, condition=is_atomic(s.x))
logic_evaluator.add_rule(contains_atomic(s.x, s.y), s.z, contains_evaluator, condition=is_function(s.x))


logic_evaluator.add_rule(pc.Xor(False,False),False);
logic_evaluator.add_rule(pc.Xor(True,False),True);
logic_evaluator.add_rule(pc.Xor(False,True),True);
logic_evaluator.add_rule(pc.Xor(True,True),False);

logic_evaluator.add_rule(pc.Not(True),False);
logic_evaluator.add_rule(pc.Not(False),True);

logic_evaluator.add_rule(pc.And(s.x,True),s.x);
logic_evaluator.add_rule(pc.And(s.x,False),False);
logic_evaluator.add_rule(pc.Or(s.x,True),True);
logic_evaluator.add_rule(pc.Or(s.x,False),s.x);

logic_evaluator.add_rule(pc.Not(pc.Not(s.x)),s.x);
logic_evaluator.add_rule(pc.And(s.x,pc.Not(s.x)),False);
logic_evaluator.add_rule(pc.Or(s.x,pc.Not(s.x)),True);

logic_evaluator.add_rule(pc.And(s.x,s.x),s.x);
logic_evaluator.add_rule(pc.Or(s.x,s.x),s.x);

logic_evaluator.add_rule(pc.equal(s.x,s.x),True);


logic_evaluator.add_rule(-s.x<-s.y,s.x<s.y)

'''
logic_evaluator.add_rule(s.x<s.x,False);
logic_evaluator.add_rule(-s.x<-s.y,s.y<s.x);
logic_evaluator.add_rule(s.x<-s.x,False);
logic_evaluator.add_rule(-s.x<s.x,False);
logic_evaluator.add_rule(s.x<=s.x,True);

logic_evaluator.add_rule(s.x>s.y,s.y<s.x);
logic_evaluator.add_rule(s.x>=s.y,s.y<=s.x);
logic_evaluator.add_rule(s.x<=s.y,pc.Or(s.x<s.y,equal(s.x,s.y)));

logic_evaluator.add_rule(And(x<y,y<x),False);

logic_evaluator.add_rule(x<oo,True);
logic_evaluator.add_rule(-oo<x,True);
logic_evaluator.add_rule(oo<x,False);
logic_evaluator.add_rule(x<-oo,False);

logic_evaluator.add_rule(And(equal(x,y),f(x)),And(equal(x,y),f(y)));
logic_evaluator.add_rule(And(Or(x,y),z),Or(And(x,z),And(y,z)));

logic_evaluator.add_rule(abs(x)<y,And(x<y,-x<y));
logic_evaluator.add_rule(x<abs(y),Or(x<y,x<-y));
logic_evaluator.add_rule(equal(abs(x),y),Or(equal(x,y),equal(x,-y)));
'''





import expresso.pycas as pc
import rule_symbols as s

evaluator = pc.RewriteEvaluator(recursive=True,split_binary=True)
canonical_form = evaluator

def normalize_exponentiation(m):
    f = m[s.x]
    if f.function != pc.multiplication:
        return False

    candidates = [arg for arg in f.args if arg.function == pc.exponentiation]
    candidates += [arg for arg in f.args if isinstance(arg.value, pc.Number) or arg == pc.I]

    if len(candidates) == 0:
        return False

    e = m[s.y]
    exponents = [(arg.args[0]**arg.args[1])**e for arg in f.args if arg.function == pc.exponentiation]
    exponents += [arg ** e for arg in f.args if isinstance(arg.value, pc.Number) or arg == pc.I]

    if len(candidates) != len(f.args):
        exponents += [pc.multiplication(*[arg for arg in f.args if arg not in candidates])**e]

    m[s.z] = pc.multiplication(*exponents)

canonical_form.add_rule(pc.multiplication(s.x),s.x)
canonical_form.add_rule(pc.addition(s.x),s.x)
canonical_form.add_rule(pc.exponentiation(s.x),s.x)

canonical_form.add_rule(1/s.x, s.x**-1)
canonical_form.add_rule(pc.exp(s.x), pc.e**s.x)
canonical_form.add_rule(s.x**s.y, s.z, normalize_exponentiation,condition=pc.equal(pc.DominantType(pc.Type(s.y),pc.Types.Real),pc.Types.Real))

canonical_form.add_rule(pc.exp(s.x),pc.e**s.x)
canonical_form.add_rule(pc.sqrt(s.x),s.x**(1/pc.S(2)))


canonical_form.add_rule(s.x>s.y,s.y<s.x)
canonical_form.add_rule(s.x>=s.y,s.y<=s.x)
canonical_form.add_rule(s.x<=s.y,pc.Or(s.x<s.y,pc.equal(s.x,s.y)))
canonical_form.add_rule(pc.unequal(s.x,s.y),pc.Not(pc.equal(s.x,s.y)));


canonical_form.add_rule(abs(s.x),pc.Max(s.x,-s.x),condition=pc.equal(pc.DominantType(pc.Type(s.x),pc.Types.Real),pc.Types.Real))

canonical_form.add_rule(pc.Max(s.a,s.b),-pc.Min(-s.a,-s.b))

canonical_form.add_rule(pc.tan(s.x),pc.sin(s.x)/pc.cos(s.x))
canonical_form.add_rule(pc.cot(s.x),pc.cos(s.x)/pc.sin(s.x))


#canonical_form.add_rule(abs(s.x)<s.y,pc.And(s.x<s.y,-s.x<s.y),condition=pc.And(s.y>0,pc.equal(pc.DominantType(pc.Type(s.x),pc.Types.Real),pc.Types.Real)))

from .type_evaluator import issubtype

format_evaluator = pc.RewriteEvaluator(recursive=True,split_binary=True)
format_evaluator.add_rule(s.x**-1,1/s.x)
format_evaluator.add_rule(s.x ** -s.y, 1 / s.x ** s.y, lambda m:isinstance(m[s.y].value, pc.Number))
format_evaluator.add_rule(pc.e**s.x,pc.exp(s.x))
format_evaluator.add_rule(s.x**(1/pc.S(2)),pc.sqrt(s.x))


format_evaluator.add_rule(pc.sqrt(s.x)*pc.sqrt(s.y),pc.sqrt(s.x*s.y),condition=pc.Or(s.x>0,s.y>0))
format_evaluator.add_rule(pc.sqrt(s.x)/pc.sqrt(s.y),pc.sqrt(s.x/s.y),condition=pc.Or(s.x>0,s.y>0))


format_evaluator.add_rule(s.x**s.a/s.y**s.a,(s.x/s.y)**s.a,condition=pc.Or(s.x>0,s.y>0,issubtype(s.a,pc.Types.Integer)))
format_evaluator.add_rule(s.x**s.a*s.y**s.a,(s.x*s.y)**s.a,condition=pc.Or(s.x>0,s.y>0,issubtype(s.a,pc.Types.Integer)))
format_evaluator.add_rule(s.x**s.a*s.y**-s.a,(s.x/s.y)**s.a,condition=pc.Or(s.x>0,s.y>0,issubtype(s.a,pc.Types.Integer)))

format_evaluator.add_rule(pc.Min(s.a,-s.a),-abs(s.a))
format_evaluator.add_rule(pc.Min(-s.a,s.a),-abs(s.a))
format_evaluator.add_rule(pc.Min(s.a,s.b),-abs(s.a),condition=pc.equal(-s.b,s.a))
format_evaluator.add_rule(-(-s.a),s.a)

format_evaluator.add_rule(pc.equal(s.a,s.a),True)
format_evaluator.add_rule(-(s.a+s.b),-s.a-s.b)

format_evaluator.add_rule(pc.sin(s.x)/pc.cos(s.x),pc.tan(s.x))
format_evaluator.add_rule(pc.cos(s.x)/pc.sin(s.x),pc.cot(s.x))

format_evaluator.add_rule((-s.a)*s.b,-(s.a*s.b))
format_evaluator.add_rule(pc.Not(pc.equal(s.x,s.y)),pc.unequal(s.x,s.y));

format_evaluator.add_rule(pc.Or(s.x<s.y,pc.equal(s.x,s.y)),s.x<=s.y)

format_evaluator.add_rule((s.x + pc.conjugate(s.x)),2*pc.real(s.x))
format_evaluator.add_rule((s.x - pc.conjugate(s.x)),2j*pc.imag(s.x))
format_evaluator.add_rule(s.x/s.x,1)
format_evaluator.add_rule(1*s.x,s.x)
format_evaluator.add_rule(s.x/1,s.x)
format_evaluator.add_rule(-1*s.x,-s.x)
format_evaluator.add_rule(-1*-1,1)
format_evaluator.add_rule(pc.I*pc.I,-1)







from .expression import Function,BinaryOperator,UnaryOperator,Symbol,Wildcard,Number,Expression,S
from .expression import One,Zero,NaN,I,addition,negative,multiplication,fraction,exponentiation,addition_group,multiplication_group,real_field,complex_field,Or,And,Xor,Not,mod,equal,unequal,In,NotIn,Less,LessEqual,Greater,GreaterEqual,Abs,Tuple

import expresso
import math

# Symbols
# -------

class SymbolicConstant(object):
    def __init__(self,name):
        self.name = name

def create_symbolic_constant(name):
    return S(expresso.create_object(SymbolicConstant(name),'symbolic constant %s' % name))

e = create_symbolic_constant('e')
pi = create_symbolic_constant('pi')
oo = create_symbolic_constant('infinity')

# Various Functions
# -----------------

factorial = UnaryOperator('!',expresso.postfix,0)

sign = Function('sign',argc = 1)
floor = Function('floor',argc = 1)
ceil = Function('ceil',argc = 1)
round = Function('round',argc = 1)

Max = Function('max',argc = 2)
Min = Function('min',argc = 2)

real = Function('real',argc = 1)
imag = Function('imag',argc = 1)
conjugate = Function('conjugate',argc = 1)

Indicator = Function('indicator',argc = 1)

InnerPiecewise = BinaryOperator('}{', expresso.associative, expresso.non_commutative, -14)
OuterPiecewise = Function('outer piecewise')

def piecewise(*args):
    return OuterPiecewise(InnerPiecewise(*args))

# Calculus
# --------

derivative = Function('derivative',argc = 2)
evaluated_at = Function('evaluated_at',argc = 3)
tmp = UnaryOperator('tmp_',expresso.prefix,0)

# Trigonometric Functions
# -----------------------

sqrt = Function('sqrt',argc = 1)
exp = Function('exp',argc = 1)
log = Function('log',argc = 1)
sin = Function('sin',argc = 1)
cos = Function('cos',argc = 1)
asin = Function('asin',argc = 1)
acos = Function('acos',argc = 1)
tan = Function('tan',argc = 1)
atan = Function('atan',argc = 1)
atan2 = Function('atan2',argc = 2)
cot = Function('cot',argc = 1)
acot = Function('acot',argc = 1)

sinh = Function('sinh',argc = 1)
cosh = Function('cosh',argc = 1)
asinh = Function('asinh',argc = 1)
acosh = Function('acosh',argc = 1)
tanh = Function('tanh',argc = 1)
atanh = Function('atanh',argc = 1)
coth = Function('coth',argc = 1)
acoth = Function('acoth',argc = 1)

# Type
# ----

Type = Function('Type',argc = 1)
DominantType = BinaryOperator('<>',expresso.left_associative,expresso.commutative,0)
OperationType = Function('OperationType',argc = 1)


class TypeInfo(object):
    def __init__(self,**kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        return repr('pyCAS.TypeInfo<%s>' % self.__dict__)


inverse_python_types = {}

def create_type(name,**kwargs):
    res = S(expresso.create_object(TypeInfo(name=name,**kwargs),'pyCAS type ' + name))
    if 'python_type' in kwargs:
        inverse_python_types[kwargs['python_type']] = res
    return res

class Types:
  Boolean = create_type('Boolean',python_type=bool,c_type='bool')
  Natural = create_type('Natural',python_type=long,c_type='unsigned')
  Integer = create_type('Integer',python_type=long,c_type='long')
  Rational = create_type('Rational',python_type=float,c_type='double')
  Real = create_type('Real',python_type=float,c_type='double')
  Complex = create_type('Complex',python_type=complex,c_type='c_complex')
  Imaginary = create_type('Imaginary',python_type=complex,c_type='c_complex')
  Unit = create_type('Unit')
  Type = create_type('Type')


def is_in_type(expression,type):
    return equal(DominantType(Type(expression),type),type)

# Custom Function
# ---------------

CustomFunction = Function("CustomFunction")

def custom_function(name,argc = None,return_type = None,**kwargs):
    
    class CustomFunctionData(object):
        def __init__(self,**kwargs):
            self.__dict__.update(kwargs)

    func_obj = expresso.create_object(CustomFunctionData(name=name,**kwargs),name)
    
    if argc is not None and not isinstance(argc,(tuple,list)):
        argc = [argc]
        
    class CustomFunctionDelegate(object):
        
        def __init__(self,func_obj,name,argc):
            self.func_obj = func_obj
            self.name = name
            self.argc = argc
        
        def __repr__(self):
            return name
            
        def __call__(self,*args):
            if self.argc != None and len(args) not in self.argc:
                raise ValueError('%s takes %s arguments' % (self.name,' or '.join([str(s) for s in self.argc])))
            args = [self.func_obj] + list(args)
            return CustomFunction(*args)

    res = CustomFunctionDelegate(func_obj,name,argc)

    if return_type != None:
        if not argc:
            raise ValueError('argc needs to be defined to register result type')

        # TODO: register in local context, not global evaluators
        from evaluators.type_evaluator import evaluator
        for c in argc:
            evaluator.add_rule(Type(res(*[Wildcard(str(s)) for s in range(c)])),return_type)

    return res


ArrayAccess = Function("ArrayAccess")


class type_converters:

    import numpy as np

    numpy_c_typenames = {
        np.bool.__name__:'bool',
        np.int.__name__:'int',
        np.int8.__name__:'int8_t',
        np.int16.__name__:'int16_t',
        np.int32.__name__:'int32_t',
        np.int64.__name__:'int64_t',
        np.float32.__name__:'float',
        np.float.__name__:'double',
        np.float64.__name__:'double',
        np.complex:'complex<double>',
        np.complex128.__name__:'complex<double>'
    }


def array(name,inarray):
    import numpy as np

    if not inarray.dtype.name in type_converters.numpy_c_typenames:
        raise ValueError('Array dtype must be one of the following: %s' % ','.join(type_converters.numpy_c_typenames.keys()))

    array = inarray #np.ascontiguousarray(inarray,dtype=cast_type)
    pointer = array.ctypes.data

    class ArrayAccessDelegate(object):

        def __init__(self,name,array):
            self.name = name
            self.array_obj = expresso.create_object(array,'%s__id_%r' % (name,id(array)))
            self.argc = len(array.shape)

        def __repr__(self):
            return self.name

        def __call__(self,*args):
            if len(args) != self.argc:
                raise ValueError('%s takes %s arguments' % (self.name,self.argc))
            args = [self.array_obj] + list(args)
            return ArrayAccess(*args)

    for t,v in inverse_python_types.iteritems():
        if isinstance(inarray.dtype,t):
            # TODO: register in local context, not global evaluators
            from evaluators.type_evaluator import evaluator
            evaluator.add_rule(ArrayAccessDelegate([Wildcard(str(i)) for i in range(array.shape)]),v)
            break

    return ArrayAccessDelegate(name,array)

unfoldable = Function('unfoldable',argc=1)

def parameter(name,value):
    import numpy as np
    value = np.array([value])

    def set_value(new_value):
        value[0] = new_value
    def get_value():
        return value[0]

    svalue = array(name,value)(unfoldable(0))

    from .expression import latex,printer
    @latex.register_target(svalue)
    @printer.register_target(svalue)
    def print_value(printer,expr):
        return printer.format_name(name)

    svalue.set_value = set_value
    svalue.get_value = get_value

    return svalue























from .lambda_compiler import LambdaCompiler,visitor
import numpy as np
from .. import expression as e
from .. import functions as f
from mpmath import mp

class NumpyCompiler(LambdaCompiler):

    def __init__(self):
        super(NumpyCompiler,self).__init__()

    @visitor.on('expr',parent = LambdaCompiler)
    def visit(self,expr):
        raise ValueError('cannot compile expression: %s' % expr)

    def get_function(self,name):
        func = None
        if name in np.__dict__:
            return np.__dict__[name]
        if name[0] == 'a':
            arcname = 'arc' + name[1:]
            if arcname in np.__dict__:
                return np.__dict__[arcname]
        return None

    @visitor.function(f.InnerPiecewise)
    def visit(self,expr):

        restype = f.Type(expr).evaluate(cache=self.cache).value
        if isinstance(restype ,f.TypeInfo):
            ptype = restype .__dict__.get('python_type')
        else:
            ptype = None

        cond_args = [self.visit(arg.args[1]) for arg in expr.args ]
        eval_args = [self.visit(arg.args[0]) for arg in expr.args ]

        def evaluate(args):

            #dtype = ptype if ptype is not None else args['_dtype']
            #shape = args['_shape']
            #res = np.zeros(shape,dtype = dtype)
            #unset = np.ones(shape,dtype = bool)

            is_arr = False

            for cond,val in zip(cond_args,eval_args):

                valid = cond(args)

                if not isinstance(valid,np.ndarray) or valid.shape == [1]:
                    if valid == False:
                        continue
                    if valid == True:
                        if not is_arr:
                            return val(args)
                        valid = unset
                else:
                    if not is_arr:
                        if np.all(valid):
                            return val(args)
                        if np.all(valid==False):
                            continue
                        shape = valid.shape
                        is_arr = True
                        unset = np.ones(shape,dtype = bool)
                        if ptype:
                            res = np.zeros(shape,dtype = ptype)
                        else:
                            res = np.zeros(shape)

                    valid &= unset

                new_args = { name:arg[valid] if isinstance(arg,np.ndarray) and arg.shape==shape else arg
                            for name,arg in args.iteritems() }

                values = np.array(val(new_args))

                if not np.can_cast(values.dtype,res.dtype):
                    res = res.astype(values.dtype)

                res[valid] = values

                unset &= np.logical_not(valid)

            return res

        return evaluate

    @visitor.obj(e.Number)
    def visit(self,expr):
        v = expr.value
        return lambda args:v

    @visitor.function(f.Not)
    def visit(self,expr):
        arg = self.visit(expr.args[0])
        return lambda args:np.logical_not( arg(args) )

    @visitor.function(f.Max)
    def visit(self,expr):
        arguments = [self.visit(arg) for arg in expr.args]
        return lambda args:np.maximum( *[arg(args) for arg in arguments] )

    @visitor.function(f.Min)
    def visit(self,expr):
        arguments = [self.visit(arg) for arg in expr.args]
        return lambda args:np.minimum( *[arg(args) for arg in arguments] )

    @visitor.function(f.ArrayAccess)
    def visit(self,expr):
        array = expr.args[0].value

        indices = [self.visit(arg) for arg in expr.args[1:]]

        def access_function(args):

            idx = [arg(args) for arg in indices[::-1]]

            shape = None
            for i in idx:
                if isinstance(i,np.ndarray):
                    shape = i.shape
                    break

            if shape != None:
                idx = [arg.astype(int) if isinstance(arg,np.ndarray) else int(arg)*np.ones(shape,dtype=int)
                       for arg in idx]
            else:
                idx = [int(arg) for arg in idx]

            valid =  reduce(np.logical_and,[ (i >= 0) & (i<s) for i,s in zip(idx,array.shape) ])

            if np.all(valid):
                return array[tuple(idx)]
            if np.all(valid == False):
                return self.value_converter(0)

            res = np.zeros(valid.shape,dtype = array.dtype)
            idx = [i[valid] for i in idx]
            res[valid] = array[tuple(idx)]

            return res

        return access_function

    @visitor.obj(mp.mpc)
    def visit(self,expr):
        return lambda args:complex(expr.value)

    @visitor.obj(mp.mpf)
    def visit(self,expr):
        return lambda args:float(expr.value)


def get_example_arg(args):
    for arg in args.values():
        if isinstance(arg,np.ndarray):
            return arg
    for arg in args.values():
        if hasattr(arg,"__len__"):
            return arg
    return args.values()[0]


def prepare_arguments(args):
    example_arg = get_example_arg(args)

    if not hasattr(example_arg,"__len__"):
        shape = None
        args = { name:np.array([arg]) for name,arg in args.iteritems() }
    else:
        shape = np.array(example_arg).shape
        args = { name:(np.array(arg) if hasattr(arg,"__len__") else arg) for name,arg in args.iteritems() }

    return args,shape


def make_parallel(f):

    import threading
    from multiprocessing import cpu_count

    def run_parallel_thread(_processes = cpu_count(),**args):

        args,shape = prepare_arguments(args)
        size = shape[0] if shape else 1
        _processes = min(size,_processes)

        if _processes == 1:
            return f(**args)

        step = int(size/_processes)
        slices = [[i*step,(i+1)*step] for i in range(_processes)]
        slices[-1][1] = size

        result = np.zeros(shape,dtype = f.restype)

        def worker(s,args):
            args = {name:(value[s[0]:s[1]]) if isinstance(value,np.ndarray) and value.shape==shape else value for name,value in args.iteritems()}
            args['_slice'] = s
            result[s[0]:s[1]] = f(**args)

        threads = []

        for s in slices:
            t = threading.Thread(target=worker,args=[s,args])
            threads.append(t)
            t.start()

        for t in threads:
            t.join()


        return result

    return run_parallel_thread

def numpyfy(expr,parallel = False,restype = None):

    from expresso.pycas.evaluators.optimizers import optimize_for_compilation

    compiler = NumpyCompiler()
    res = compiler.visit(optimize_for_compilation(e.S(expr)))

    if restype is None:
        restype = f.Type(expr).evaluate(cache=compiler.cache).value
        if isinstance(restype,f.TypeInfo):
            restype = restype.__dict__.get('python_type')
            if restype == None:
                restype = complex
        else:
            restype = complex

    def call(**args):

        args,shape = prepare_arguments(args)
        cres = np.array(res(args)).astype(restype)

        if not shape:
            if cres.shape:
                cres = cres[0]
        else:
            if shape and cres.shape != shape:
                cres = np.ones(shape)*cres

        return cres

    call.restype = restype

    if parallel:
        return make_parallel(call)
    else:
        return call


def ncompile(*function_definitions):

    functions = {}

    for definition in function_definitions:

        if definition.return_type is not None:
            restype = definition.return_type.value.__dict__.get('python_type')
        else:
            restype = None

        f = numpyfy(definition.expr,parallel=definition.parallel,restype=restype)

        if definition.arg_types:
            # TODO: implement argument type conversions
            arg_types = [arg.value.__dict__.get('python_type') for arg in definition.arg_types]

        arg_names = [arg.name for arg in definition.args]

        class Delegate(object):
            def __init__(self,f,arg_names):
                self.f = f
                self.arg_names = arg_names
            def __call__(self, *args, **kwargs):
                args = { n:a for n,a in zip(self.arg_names,args) }
                res = kwargs.pop('res',None)
                if res is not None:
                    res[:] = self.f(**args)
                    return res
                else:
                    return self.f(**args)

        functions[definition.name] = Delegate(f,arg_names)

    class lib(object):
        def __init__(self,functions):
            self.__dict__.update(functions)

    return lib(functions)














import _expresso as core
from _expresso import associative,left_associative,right_associative,non_associative,commutative,non_commutative,postfix,prefix

class Expression(core.Expression):

    def __init__(self,expr,S):
        if isinstance(expr,(core.Expression,Expression)):
            super(Expression,self).__init__(expr)
        else:
            super(Expression,self).__init__(S(expr))
        self.S = S
    
    @property
    def function(self):
        if not hasattr(self,'_function'):
            f = super(Expression,self).function()
            if f is not None:
                F = Function(f,S = self.S)
            else:
                F = None
        return F
    
    @property
    def is_function(self):
        return super(Expression,self).is_function()    

    @property
    def is_symbol(self):
        return super(Expression,self).is_symbol()

    @property
    def is_atomic(self):
        return super(Expression,self).is_atomic()

    @property
    def is_wildcard_symbol(self):
        return super(Expression,self).is_wildcard_symbol()

    @property
    def is_wildcard_function(self):
        return super(Expression,self).is_wildcard_function()

    @property
    def is_wildcard(self):
        if self.is_function:
            return self.is_wildcard_function
        return self.is_wildcard_symbol

    @property
    def value(self):
        return self.get_value()

    @property
    def args(self):
        args = self.get_arguments()
        if args == None:
            return []
        return [self.S(arg) for arg in args]

    @property
    def name(self):
        if self.is_function:
            return self.function.name
        return super(Expression,self).__repr__()

    def get(self,type):
        v = self.value
        if isinstance(v,type):
            return v
        return None

    def replace(self,*args):
        if len(args) == 1:
            if isinstance(args[0],ReplacementMap):
                rep = args[0]._replacement_map
            else:
                rep = ReplacementMap(args[0],self.S)._replacement_map
        elif isinstance(args[0],(list,tuple)):
            if len(args) == 1 and isinstance(args[0][0],(list,tuple)):
                rep = ReplacementMap(args[0],self.S)._replacement_map
            else:
                rep = ReplacementMap(args,self.S)._replacement_map
        elif len(args) == 2:
            rep = core.replacement_map()
            rep[self.S(args[0])] = self.S(args[1])
        else:
            raise ValueError('invalid substitution arguments')

        #from evaluator import ReplaceEvaluator
        #return ReplaceEvaluator(rep,S=self.S)(self)

        return self.S(core.replace(self,rep))


    def match(self,search):
        for expr in core.commutative_permutations(search):
            res = core.match(self,expr)
            if res:
                return ReplacementMap(core.match(self,expr),self.S)
        return None

def WrappedType(T,**parameters):
    
    class Wrapped(T):
        
        @staticmethod
        def _get_wrapped_parameter(name):
            return parameters.get(name)
        
        def __init__(self,*args,**kwargs):
            kwargs.update(parameters)
            super(Wrapped,self).__init__(*args,**kwargs)
    
    Wrapped.__name__ = "Wrapped" + T.__name__
    
    return Wrapped

WrappedExpression = lambda S:WrappedType(Expression,S=S)

class Function(object):

    def __init__(self,F,argc = None,S = None):
        if S == None:
            raise ValueError('no parameter S provided')
        self.S = S
        if argc is not None:
            if isinstance(argc,int):
                argc = [argc]
            self.argc = argc
        self._function = F

    def __call__(self,*args):
        if hasattr(self,'argc') and len(args) not in self.argc:
            raise ValueError("%s takes %s arguments" % (str(self),' or '.join([str(v) for v in self.argc])))
        return self.S(core.create_call(self._function,[self.S(arg) for arg in args]))

    def __repr__(self):
        return self.name

    @property
    def name(self):
        return self._function.get_name()

    @property
    def symbol(self):
        return self._function.get_symbol()

    @property
    def is_commutative(self):
        return self._function.is_commutative()

    @property
    def is_associative(self):
        return self._function.is_associative()

    @property
    def is_prefix(self):
        return self._function.is_prefix()

    @property
    def is_postfix(self):
        return self._function.is_postfix()

    @property
    def is_operator(self):
        return isinstance(self._function,core.Operator)    

    @property
    def is_unary_operator(self):
        return isinstance(self._function,core.UnaryOperator)    

    @property
    def is_binary_operator(self):
        return isinstance(self._function,core.BinaryOperator)    

    @property
    def precedence(self):
        return self._function.get_precedence()

    def __hash__(self):
        return self.name.__hash__()

    def __eq__(self,other):
        if not isinstance(other,Function):
            return False
        return self.name == other.name

    def __ne__(self,other):
        return not self == other

    
def WrappedFunction(F,S,argc = None):
    
    if argc == None:
        class WrappedFunction(Function):
            def __init__(self,*args,**kwargs):
                argc = kwargs.pop('argc',None)
                super(WrappedFunction,self).__init__(F(*args,**kwargs),argc,S)
    else:
        class WrappedFunction(Function):
            def __init__(self,*args,**kwargs):
                super(WrappedFunction,self).__init__(F(*args,**kwargs),argc,S)

    WrappedFunction.__name__ = "Wrapped" + F.__name__
                
    return WrappedFunction


class ReplacementMap(object):

    def __init__(self,rep = None,S = None):
        
        if isinstance(rep,core.replacement_map):
            self._replacement_map = rep
        elif rep != None:
            self._replacement_map = core.replacement_map()
            for key, value in dict(rep).iteritems():
                self._replacement_map[S(key)] = S(value)
        else:
            self._replacement_map = core.replacement_map()

        self.S = S
        
    def __str__(self):
        return str(dict(self))
    
    def __iter__(self):
        def generator():
            for v in self._replacement_map:
                yield (self.S(v.key()),self.S(v.data()))
        return generator()
    
    def __getitem__(self,key):
        return self.S(self._replacement_map[self.S(key)])

    def __setitem__(self,key,value):
        self._replacement_map[self.S(key)] = self.S(value)



WrappedReplacementMap = lambda S:WrappedType(ReplacementMap,S=S)

class Group(object):

    def __init__(self,operation,inverse = None,neutral = None,S = None):
        if isinstance(operation,core.Group):
            self._group = operation
        else:
            self._group = core.Group(operation._function,inverse._function,S(neutral))
        self.S = S
        
    @property
    def operation(self):
        return Function(self._group.get_operation(),self.S)
    @property
    def inverse(self):
        return Function(self._group.get_inverse(),self.S)
    @property
    def neutral(self):
        return self.S(self._group.neutral)
    
    def __repr__(self):
        return "Group" + str((self.operation,self.inverse,self.neutral))
    
WrappedGroup = lambda S:WrappedType(Group,S=S)

class Field(object):

    def __init__(self,additive_group,multiplicative_group,S):
        self._field = core.Field(additive_group._group,multiplicative_group._group)
        self.S = S
        
    @property
    def additive_group(self):
        return Group(self._field.additive_group,self.S)
    
    @property
    def multiplicative_group(self):
        return Group(self._field.multiplicative_group,self.S)

        
WrappedField = lambda S:WrappedType(Field,S=S)

class MulplicityList(object):

    def __init__(self,arg,operation_group = None,mulplicity_function = None,field = None,S = None):
        if S == None:
            raise ValueError("Parameter S undefined")
        if arg == None:
            self._mlist = core.MulplicityList(operation_group._group,mulplicity_function._function,field._field)
        elif isinstance(arg,core.MulplicityList):
            self._mlist = arg
        else:
            self._mlist = core.MulplicityList(S(arg),operation_group._group,mulplicity_function._function,field._field)
        self.S = S
    
    def __len__(self):
        return len(self._mlist)
    
    def __iter__(self):
        def generator():
            for v in self._mlist:
                yield (self.S(v.value),self.S(v.mulplicity))
        return generator()
    
    def __getitem__(self,index):
        if index >= len(self._mlist):
            raise ValueError("index out of range")
        v = self._mlist[index]
        return (self.S(v.value),self.S(v.mulplicity))

    def __repr__(self):
        return str(list(self))
    
    def as_expression(self):
        return self.S(self._mlist.as_expression())
    
    def _wrap(self,mlist):
        return MulplicityList(mlist,S = self.S)
    
    def intersection(self,other,get_inner = None):
        if get_inner:
            return self._wrap(self._mlist.intersection(other._mlist,lambda a,b:get_inner(self.S(a),self.S(b))))
        return self._wrap(self._mlist.intersection(other._mlist))
    
    def sub(self,other):
        return self._wrap(self._mlist.difference(other._mlist))

    def sum(self,other):
        return self._wrap(self._mlist.sum(other._mlist))

    def pow(self,expr):
        return self._wrap(self._mlist.power(self.S(expr)))

    def __sub__(self,other):
        return self.sub(other)

    def __add__(self,other):
        return self.sum(other)

    def __pow__(self,other):
        return self.pow(other)

WrappedMulplicityList = lambda S:WrappedType(MulplicityList,S=S)
    
def wrapped_postorder_traversal(S):
    def postorder_traversal(expr):
        for expr in core.postorder_traversal(expr):
            yield S(expr)
    return postorder_traversal


def wrapped_preorder_traversal(S):
    def preorder_traversal(expr):
        for expr in core.preorder_traversal(expr):
            yield S(expr)
    return preorder_traversal


def wrapped_commutative_permutations(S):
    def commutative_permutations(expr):
        for expr in core.commutative_permutations(expr):
            yield S(expr)
    return commutative_permutations





from expression import core,WrappedExpression
import expression
import evaluator
import printer
import visitor

from expression import associative,non_associative,left_associative,right_associative,commutative,non_commutative,postfix,prefix

def create_object(data,unique_name=None):
    '''
Create an atomic expression that holds data as the value.
Note that two objects will be treated the the same if they
have the same 'unique_name' (by default given by the type and
rerp(data)).
    '''
    if unique_name is None:
        unique_name = '%s:(%s)' % (type(data).__name__, repr(data))
    return core.create_object(data,unique_name)

def create_symbol(name):
    return core.create_symbol(name)

class WrappedExpressionTypes(object):

    def __init__(self,Expression):
        
        S = Expression._get_wrapped_parameter('S')

        self.Function = expression.WrappedFunction(core.Function,S)
        self.BinaryOperator = expression.WrappedFunction(core.BinaryOperator,S)
        self.UnaryOperator = expression.WrappedFunction(core.UnaryOperator,S,argc = 1)
        
        self.Group = expression.WrappedGroup(S)
        self.Field = expression.WrappedField(S)
        self.ReplacementMap = expression.WrappedReplacementMap(S)
        self.MulplicityList = expression.WrappedMulplicityList(S)

        self.MatchCondition = evaluator.WrappedMatchCondition(S)
        self.Rule = evaluator.WrappedRule(S)
        self.RewriteEvaluator = evaluator.WrappedRewriteEvaluator(S)
        self.MultiEvaluator = evaluator.WrappedMultiEvaluator(S)
        self.StepEvaluator = evaluator.WrappedStepEvaluator(S)
        self.ReplaceEvaluator = evaluator.WrappedReplaceEvaluator(S)

        self.postorder_traversal = expression.wrapped_postorder_traversal(S)
        self.preorder_traversal = expression.wrapped_preorder_traversal(S)    
        self.commutative_permutations = expression.wrapped_commutative_permutations(S) 
        
def default_converter(expr):    
    if isinstance(expr,core.Expression):
        return Expression(expr)
    elif isinstance(expr,Expression):
        return expr
    elif isinstance(expr,str):
        return Expression(create_symbol(expr))
    else:
        return Expression(create_object(expr))
    
def comp(a,b):
    return core.Expression.__lt__(a,b)
    
#Expression = WrappedExpression(default_converter)
#locals().update(WrappedExpressionTypes(Expression).__dict__)





from expression import *
from functions import *
from expresso.visitor import add_target,add_target_obj

@add_target(printer,addition)
@add_target(latex,addition)
def visit(printer,expr):
    neg_args = [arg for arg in expr.args if arg.function == negative]
    covered = set(neg_args)
    rest = [arg for  arg in expr.args if arg not in covered]
    rest_str = '+'.join(printer.printed_operator_arguments(expr,rest))
    if len(neg_args) == 0:
        return rest_str
    neg_str = '-'.join(printer.printed_operator_arguments(expr,[arg.args[0] for arg in neg_args]))
    return rest_str + '-' + neg_str

@add_target(latex,multiplication)
def visit(printer,expr):
    denominators = [arg for arg in expr.args if arg.function == fraction]

    if len(denominators)>0:
        numerators = [arg for arg in expr.args if arg.is_atomic]
        if len(numerators) == 0:
            numerators = [One]

        covered = set(numerators + denominators)
        rest = [arg for  arg in expr.args if arg not in covered]

        denom_str = printer(multiplication(*[arg.args[0] for arg in denominators]))
        num_str =   printer(multiplication(*numerators))

        if len(rest) == 0:
            rest_str = ''
        elif len(rest) == 1:
            rest_str = printer(rest[0])
            if printer.needs_brackets_in(rest[0],expr):
                rest_str = printer.bracket_format() % rest_str
        else:
            rest = multiplication(*rest)
            rest_str =  printer(rest)

        return r'\frac{%s}{%s} \, %s ' % (num_str,denom_str,rest_str)

    needs_dot = lambda x: type(x.value)==Number

    need_dot = [x for x in expr.args if needs_dot(x)]
    rest     = [x for x in expr.args if not needs_dot(x)]

    res  = '\cdot '.join(printer.printed_operator_arguments(expr,need_dot))
    if len(need_dot) > 1:
        res += '\cdot '
    elif len(need_dot) == 1:
        res += '\, '
    res += '\, '.join(printer.printed_operator_arguments(expr,rest))

    return res

@add_target_obj(printer,SymbolicConstant)
@add_target_obj(latex,SymbolicConstant)
def visit(printer,expr):
    return expr.value.name

@add_target(printer,multiplication)
def visit(printer,expr):
    res = printer.print_operator_argument(expr.args[0],expr)
    for arg in expr.args[1:]:
        if arg.function == fraction:
            res += '/' + printer.print_operator_argument(arg.args[0],expr)
        else:
            res += '*' + printer.print_operator_argument(arg,expr)
    return res

@add_target(latex,fraction)
def visit(printer,expr):
    return r'\frac{1}{%s}' % printer(expr.args[0])

@add_target(latex,exponentiation)
def visit(printer,expr):
    parg = printer.printed_operator_arguments(expr,end=-1)
    parg += [printer(expr.args[-1])]
    return '^'.join(['{%s}' % arg for arg in parg])

@latex.register_target(derivative)
def visit(printer,expr):

    def flatten(expr,arguments=[]):
        arguments.append(expr.args[1])
        inner = expr.args[0]
        if inner.function == derivative:
            return flatten(inner,arguments)
        return inner,arguments
    
    inner,arguments = flatten(expr)
    
    argc = len(arguments)
    mul_args = [(arguments[0],1)]
    for arg in arguments[1:]:
        if mul_args[-1][0] == arg:
            mul_args[-1] = (arg,mul_args[-1][1]+1)
        else:
            mul_args.append((arg,1))
        
    formatted_arguments = [r'{\partial %s}' % printer(arg[0]) if arg[1] == 1 else r'{\partial{%s}^{%s}}' % (printer(arg[0]),arg[1]) for arg in mul_args]
    
    if argc == 1:
        formated_inner = '\partial %s' % printer(inner)
    else:
        formated_inner = '\partial^{%s} %s' % (argc,printer(inner))
    
    return r'\frac{%s}{%s}' % (formated_inner,'\,'.join(formatted_arguments))

@add_target(latex,evaluated_at)
def visit(printer,expr):
    return r'\left[ %s \right]_{%s = %s}' % tuple(printer(arg) for arg in expr.args)


latex.register_printer(pi,lambda p,e:r'\pi ')
latex.register_printer(e,lambda p,e:r'e ')
latex.register_printer(oo,lambda p,e:r'\infty ')

latex.register_printer(I,lambda p,e:r'i ')
printer.register_printer(I,lambda p,e:r'i')


latex.register_printer(Abs,lambda p,e: r"\left| %s \right|" % p(e.args[0]))

latex.register_printer(conjugate,lambda p,e: r"\overline{ %s }" % p(e.args[0]))


latex.register_printer(tmp,lambda p,e: p.print_postfix_operator(e,r"'"))
latex.register_printer(sqrt,lambda p,e:r"\sqrt{%s} " % p(e.args[0]))

latex.register_printer(real,lambda p,e: p.print_function(e,r"\Re "))
latex.register_printer(imag,lambda p,e: p.print_function(e,r"\Im "))

latex.register_printer(Indicator,lambda p,e: p.print_function(e,r"\, \mathbb{1} "))
latex.register_printer(Not,lambda p,e: p.print_unary_operator(e,r"\neg "))

latex.register_printer(LessEqual,lambda p,e: p.print_binary_operator(e,r"\le "))
latex.register_printer(GreaterEqual,lambda p,e: p.print_binary_operator(e,r"\ge "))
latex.register_printer(Or,lambda p,e: p.print_binary_operator(e,r"\lor "))
latex.register_printer(And,lambda p,e: p.print_binary_operator(e,r"\land "))
latex.register_printer(Xor,lambda p,e: p.print_binary_operator(e,r"\veebar "))


latex.register_printer(Tuple,lambda p,e: p.print_function(e,r""))

@add_target(latex, InnerPiecewise)
def visit(printer,expr):   
    for arg in expr.args:
        if arg.function != Tuple:
            return printer.print_function(expr,name=r"\text{piecewise} ")

    outer = r"\begin{cases} %s \end{cases}"

    s = len(expr.args)-1
    inner_list = [r"%s & \text{if } %s " % (printer(e.args[0]),printer(e.args[1]))
                         for e in expr.args[:-1]]
    e = expr.args[-1]

    if e.args[1] == S(True):
        inner_list += [ r"%s & \text{otherwise } " % printer(e.args[0])]
    else:
        inner_list += [ r"%s & \text{if } %s " % (printer(e.args[0]),printer(e.args[1])) ]

    inner = r"\\ ".join(inner_list)
    return outer % inner

@add_target(printer, OuterPiecewise )
@add_target(latex, OuterPiecewise )
def visit(printer,expr):
    return printer(expr.args[0])


@add_target(printer, InnerPiecewise)
def visit(printer,expr):
    return printer.print_function(expr,name="piecewise")

@add_target(latex,unequal)
def visit(printer,expr):
    return printer.print_binary_operator(expr,r" \neq ")

@add_target(latex,CustomFunction)
def visit(printer,expr):
    name = expr.args[0].name.replace('<',r'< ').replace('>',r' >')
    return printer(Function(name)(*expr.args[1:]))

@add_target(latex,ArrayAccess)
def visit(printer,expr):
    name = printer.format_name(expr.args[0].name.split('__id')[0])
    return r'%s \mathopen{} \left[ %s \right] \mathclose{} ' % (name,','.join([printer(arg) for arg in expr.args[1:]]))

@add_target(printer,ArrayAccess)
def visit(printer,expr):
    name = printer.format_name(expr.args[0].name.split('__id')[0])
    return r'%s[%s]' % (name,','.join([printer(arg) for arg in expr.args[1:]]))

@add_target_obj(printer, Number)
def visit(printer,expr):
    v = expr.value

    if v == 0:
        return '0'

    o = v
    exp = 0

    while (v % 10) == 0:
        v = v/10
        exp = exp + 1
    if exp > 1:
        if v != 1:
            return r"%se%s" % (v,exp)
        if v == 1:
            return r"1e%s" % exp

    return str(o)

@add_target_obj(latex, Number)
def visit(printer,expr):
    v = expr.value

    if v == 0:
        return '0'

    o = v
    exp = 0

    while (v % 10) == 0:
        v = v/10
        exp = exp + 1
    if exp > 1:
        if v != 1:
            return r"%s \cdot 10^{%s}" % (v,exp)
        if v == 1:
            return r"10^{%s}" % exp

    return str(o)

@add_target_obj(printer, TypeInfo)
@add_target_obj(latex, TypeInfo)
def visit(printer,expr):
    return printer.format_name(expr.value.name)

from mpmath import mp
@add_target_obj(printer, bool)
@add_target_obj(latex, bool)
def visit(printer,expr):
    return printer.format_name(str(expr.value))

@add_target_obj(latex, mp.mpc)
@add_target_obj(latex, mp.mpf)
def visit(printer,expr):
    return r"%s \; " % str(expr.value)

@add_target_obj(printer, mp.mpc)
@add_target_obj(printer, mp.mpf)
def visit(printer,expr):
    return r"%s" % str(expr.value)

















import expresso.pycas as pc
import rule_symbols as s

evaluator = pc.RewriteEvaluator(recursive=True, split_binary=True)
factor_evaluator = pc.RewriteEvaluator(recursive=True, split_binary=True)

evaluator.add_rule(s.x+0, s.x)
evaluator.add_rule(s.x-s.x, 0)
evaluator.add_rule(s.x+s.x, 2*s.x)

evaluator.add_rule(s.x*1, s.x)
evaluator.add_rule(s.x*0, 0)

evaluator.add_rule(s.x**1, s.x)
evaluator.add_rule(s.x**0, 1)
evaluator.add_rule(1**s.x, 1)
evaluator.add_rule(0**s.x, 0,condition=s.x>0)

factor_evaluator.add_rule(s.x*s.x, s.x**2)
evaluator.add_rule(s.x*s.x**-1, 1)

evaluator.add_rule((s.x**s.a)**s.b, s.x**(s.a*s.b),condition=pc.equal(pc.DominantType(pc.Type(s.b),pc.Types.Integer),pc.Types.Integer))

from .numeric_evaluator import is_even,is_uneven
from .type_evaluator import issubtype

evaluator.add_rule((s.x**s.a)**(s.a**-1), abs(s.x),condition=pc.And( issubtype(s.x,pc.Types.Real),is_even(s.a) ) )
evaluator.add_rule((s.x**s.a)**(s.a**-1), s.x,condition=pc.And( issubtype(s.x,pc.Types.Real),is_uneven(s.a)))


evaluator.add_rule((-s.x)**(s.a), s.x**s.a ,condition=is_even(s.a))
evaluator.add_rule((-s.x)**(s.a), -(s.x)**s.a ,condition=is_uneven(s.a))



factor_evaluator.add_rule(s.x**s.n*s.y**-s.n,(s.x*s.y**-1)**s.n,condition=pc.Or(s.y>0,issubtype(s.n,pc.Types.Integer)))


factor_evaluator.add_rule(s.x**s.a/s.y**s.a,(s.x/s.y)**s.a,condition=pc.Or(s.x>0,s.y>0,issubtype(s.a,pc.Types.Integer)))
factor_evaluator.add_rule(s.x**s.a*s.y**s.a,(s.x*s.y)**s.a,condition=pc.Or(s.x>0,s.y>0,issubtype(s.a,pc.Types.Integer)))
factor_evaluator.add_rule(s.x**s.a*s.y**-s.a,(s.x/s.y)**s.a,condition=pc.Or(s.x>0,s.y>0,issubtype(s.a,pc.Types.Integer)))


factor_evaluator.add_rule(s.x**s.n*s.x, s.x**(s.n+1))
factor_evaluator.add_rule(s.x**s.n*s.x**s.m, s.x**(s.n+s.m))


from logic_evaluator import is_explicit_natural

evaluator.add_rule(s.a**s.x*s.b**s.x, (s.a*s.b)**(s.x), condition=pc.And(pc.Not(pc.Or(is_explicit_natural(s.a),is_explicit_natural(s.b))),pc.Or(issubtype(s.x,pc.Types.Natural),s.a>0,s.b>0)))

evaluator.add_rule(-(s.x+s.y), -s.x-s.y)
evaluator.add_rule(s.x*-1, -s.x)
evaluator.add_rule(-(-s.x), s.x)
evaluator.add_rule((-s.x)*s.y, -(s.x*s.y))
evaluator.add_rule(1 / -s.x, -(1 / s.x))

evaluator.add_rule(-pc.S(0), 0)

def extract_intersection(m):

    ma = pc.MulplicityList(m[s.x],pc.multiplication_group,pc.exponentiation,pc.real_field)
    mb = pc.MulplicityList(m[s.y],pc.multiplication_group,pc.exponentiation,pc.real_field)

    common = ma.intersection(mb)
    if len(common) == 0:
        return False

    m[s.a] = (ma-common).as_expression()
    m[s.b] = (mb-common).as_expression()
    m[s.c] = common.as_expression()



factor_evaluator.add_rule(s.x+s.y, s.c*(s.a+s.b), extract_intersection)
factor_evaluator.add_rule(s.x-s.y, s.c*(s.a-s.b), extract_intersection)
factor_evaluator.add_rule(-s.x-s.y, -s.c*(s.a+s.b), extract_intersection)

def extract_comp_mul_intersection(m):

    ma = pc.MulplicityList(m[s.x],pc.multiplication_group,pc.exponentiation,pc.real_field)
    mb = pc.MulplicityList(m[s.y],pc.multiplication_group,pc.exponentiation,pc.real_field)

    common = ma.intersection(mb)
    if len(common) == 0:
        return False

    for v,mul in common:
        if v.function == pc.sign:
            return False

    m[s.a] = (ma-common).as_expression()
    m[s.b] = (mb-common).as_expression()
    m[s.c] = common.as_expression()

evaluator.add_rule(-s.x<-s.y, s.y<s.x)


evaluator.add_rule(s.x<s.y, pc.sign(s.c)*s.a<pc.sign(s.c)*s.b, extract_comp_mul_intersection)
evaluator.add_rule(s.x<-s.y, pc.sign(s.c)*s.a<-pc.sign(s.c)*s.b, extract_comp_mul_intersection)
evaluator.add_rule(-s.x<s.y, -pc.sign(s.c)*s.a<pc.sign(s.c)*s.b, extract_comp_mul_intersection)


evaluator.add_rule(pc.sign(s.a),1,condition=s.a>=0)
evaluator.add_rule(pc.sign(s.a),-1,condition=s.a<0)
evaluator.add_rule(pc.sign(s.a*s.b),pc.sign(s.a)*pc.sign(s.b))

evaluator.add_rule(pc.sign(s.a**s.n),1,condition=is_even(s.n))
evaluator.add_rule(pc.sign(s.a**s.n),pc.sign(s.a),condition=is_uneven(s.n))




from .logic_evaluator import is_function_type

def evaluate_fraction(m):
    ex,ey = m[s.x],m[s.y]**m[s.n]
    ma = pc.MulplicityList(ex,pc.multiplication_group,pc.exponentiation,pc.real_field)
    mb = pc.MulplicityList(ey,pc.multiplication_group,pc.exponentiation,pc.real_field)

    mbs = {k for k,v in mb}

    valid = False

    if not valid:
        for k,v in ma:
            if k in mbs:
                valid = True
                break
            if valid:
                break

    if valid == False:
        return False

    m[s.c] = (ma+mb).as_expression()

evaluator.add_rule(s.x*s.y**s.n, s.c, evaluate_fraction,condition=issubtype(s.n,pc.Types.Integer))


evaluator.add_rule(pc.log(pc.e), 1)
evaluator.add_rule(pc.log(1), 0)
evaluator.add_rule(pc.sin(0), 0)
evaluator.add_rule(pc.cos(0), 1)
evaluator.add_rule(pc.tan(0), 0)
evaluator.add_rule(pc.atan(0), 0)

evaluator.add_rule(pc.Indicator(True), 1)
evaluator.add_rule(pc.Indicator(False), 0)


evaluator.add_rule(pc.InnerPiecewise((s.a,True),s.x),(s.a,True))
evaluator.add_rule(pc.InnerPiecewise((s.a,False),s.x),s.x)
evaluator.add_rule(pc.InnerPiecewise(s.x,(s.a,False)),s.x)
evaluator.add_rule(pc.InnerPiecewise((s.a,s.x),(s.b,s.x)),(s.a,s.x))
evaluator.add_rule(pc.InnerPiecewise((s.x,s.a),(s.x,s.b)),(s.x,pc.Or(s.a,s.b)))
evaluator.add_rule(pc.InnerPiecewise(pc.InnerPiecewise(s.x)),pc.InnerPiecewise(s.x))

evaluator.add_rule(pc.OuterPiecewise((s.a,s.b)),s.a*pc.Indicator(s.b))


from .logic_evaluator import contains_atomic


excluded_derivatives = {pc.InnerPiecewise,pc.Tuple}

def check_if_excluded_derivative(m):
    return not m[s.y].function in excluded_derivatives

evaluator.add_rule(pc.derivative(s.x,s.x),1)
evaluator.add_rule(pc.derivative(s.y,s.x),0,check_if_excluded_derivative,condition=pc.Not(contains_atomic(s.y,s.x)));

evaluator.add_rule(pc.derivative(s.a+s.b,s.x),pc.derivative(s.a,s.x)+pc.derivative(s.b,s.x))
evaluator.add_rule(pc.derivative(s.a*s.b,s.x),pc.derivative(s.a,s.x)*s.b+pc.derivative(s.b,s.x)*s.a)
evaluator.add_rule(pc.derivative(-s.x,s.x),-1)
evaluator.add_rule(pc.derivative(1/s.x,s.x),-s.x**-2)
evaluator.add_rule(pc.derivative(pc.log(s.x),s.x),1/s.x)


evaluator.add_rule(pc.derivative(pc.sin(s.x),s.x),pc.cos(s.x))
evaluator.add_rule(pc.derivative(pc.cos(s.x),s.x),-pc.sin(s.x))

# TODO: Add assumptions to expressions: the following is only valid if x != -1,1 or x != -i,i
evaluator.add_rule(pc.derivative(pc.asin(s.x),s.x),1/pc.sqrt(1-s.x**2))
evaluator.add_rule(pc.derivative(pc.acos(s.x),s.x),-1/pc.sqrt(1-s.x**2))
evaluator.add_rule(pc.derivative(pc.atan(s.x),s.x),1/(1+s.x**2))
evaluator.add_rule(pc.derivative(pc.acot(s.x),s.x),-1/(1+s.x**2))



evaluator.add_rule(pc.derivative(s.x**s.n,s.x),s.n*s.x**(s.n-1),condition=(pc.equal(pc.Type(s.n))));
evaluator.add_rule(pc.derivative(s.a**s.b,s.x),pc.derivative(s.b*pc.log(s.a),s.x)*s.a**s.b);

evaluator.add_rule(pc.derivative(pc.OuterPiecewise(s.a),s.x),pc.OuterPiecewise(pc.derivative(s.a,s.x)))
evaluator.add_rule(pc.derivative(pc.InnerPiecewise((s.a,s.b),s.c),s.x),pc.InnerPiecewise((pc.derivative(s.a,s.x),s.b),pc.derivative(pc.InnerPiecewise(s.c),s.x)),condition=pc.Not(contains_atomic(s.b,s.x)))
evaluator.add_rule(pc.derivative(pc.InnerPiecewise((s.a,s.b)),s.x),(pc.derivative(s.a,s.x),s.b),condition=pc.Not(contains_atomic(s.b,s.x)))


evaluator.add_rule(pc.OuterPiecewise(s.a)**s.x,pc.OuterPiecewise(s.a**s.x))
evaluator.add_rule(pc.InnerPiecewise((s.a,s.b),s.c)**s.x,pc.InnerPiecewise((s.a**s.x,s.b),pc.InnerPiecewise(s.c)**s.x))
evaluator.add_rule(pc.InnerPiecewise((s.a,s.b))**s.x,(s.a**s.x,s.b))

evaluator.add_rule(pc.real(s.x),(s.x + pc.conjugate(s.x))/2)
evaluator.add_rule(pc.imag(s.x),(s.x - pc.conjugate(s.x))/2j)

evaluator.add_rule(pc.conjugate(s.x),s.x,condition=issubtype(s.x,pc.Types.Real))
evaluator.add_rule(pc.conjugate(s.x),-s.x,condition=issubtype(s.x,pc.Types.Imaginary))
evaluator.add_rule(pc.conjugate(s.x+s.y),pc.conjugate(s.x)+pc.conjugate(s.y))
evaluator.add_rule(pc.conjugate(s.x*s.y),pc.conjugate(s.x)*pc.conjugate(s.y))
evaluator.add_rule(pc.conjugate(-s.x),-pc.conjugate(s.x))
evaluator.add_rule(pc.conjugate(1/s.x),1/pc.conjugate(s.x))
evaluator.add_rule(pc.conjugate(s.x**-1),pc.conjugate(s.x)**-1)


def create_tmp_x(m):
    m[c] = pc.tmp(s.x)

evaluator.add_rule( pc.derivative(s.f(s.g(s.x)),s.x) ,
                    pc.evaluated_at( pc.derivative(s.f(pc.tmp(s.x)),pc.tmp(s.x)), pc.tmp(s.x), s.g(s.x) ) * pc.derivative(s.g(s.x),s.x));

evaluator.add_rule(pc.evaluated_at( s.f(s.x), s.x, s.c ), s.f(s.c), condition = pc.Not(is_function_type(s.f(s.x),pc.derivative)) )

evaluator.add_rule(pc.Min(s.a,s.b),s.a,condition=s.a<=s.b)
evaluator.add_rule(pc.Min(s.a,s.b),s.b,condition=s.b<=s.a)


from .canonical_form import canonical_form
from .logic_evaluator import logic_evaluator
from .numeric_evaluator import numeric_evaluator
from .type_evaluator import type_evaluator


main_evaluator = pc.MultiEvaluator(recursive = True, split_binary=True)
main_evaluator.add_evaluator(canonical_form)
main_evaluator.add_evaluator(numeric_evaluator)
main_evaluator.add_evaluator(evaluator)
main_evaluator.add_evaluator(factor_evaluator)
main_evaluator.add_evaluator(type_evaluator)
main_evaluator.add_evaluator(logic_evaluator)















import expresso as ex
from .printer import latex

def S(expr):
    if isinstance(expr,ex.core.Expression):
        return Expression(expr)
    if isinstance(expr,Expression):
        return expr
    if isinstance(expr,str):
        return Expression(ex.core.create_symbol(expr))
    if isinstance(expr,int):
        if expr >= 0:
            return Expression(ex.core.create_object(expr,str(expr)))
        else:
            expr = abs(expr)
            return negative(ex.core.create_object(expr,str(expr)))

    raise ValueError('Unsupported expression type: %s' % type(expr))

def W(name):
    return S(ex.core.create_wildcard_symbol(name))

def WF(name):
    return ex.Function(ex.core.WildcardFunction(name),S)

class Expression(ex.WrappedExpression(S)):
    
    def __add__(self, other):
        return addition(self,self.S(other))

    def __radd__(self, other):
        return addition(self.S(other),self)

    def __neg__(self):
        return negative(self)

    def __pos__(self):
        return self
    
    def __sub__(self, other):
        return addition(self, negative(self.S(other)))

    def __rsub__(self, other):
        return addition(self.S(other), negative(self))

    def __mul__(self, other):
        return multiplication(self,self.S(other))

    def __rmul__(self, other):
        return multiplication(self.S(other),self)

    def __div__(self, other):
        return multiplication(self, fraction(self.S(other)))

    def __rdiv__(self, other):
        other = self.S(other)
        if other == One:
            return fraction(self)
        return multiplication(self.S(other), fraction(self))
    
    def __pow__(self,other):
        return exponentiation(self,self.S(other))

    def __rpow__(self,other):
        return exponentiation(self.S(other),self)

    def _repr_latex_(self):
         return "$$%s$$" % latex(self)

One = S(1)
Zero = S(0)

Function = ex.WrappedFunction(ex.core.Function,S)
BinaryOperator = ex.WrappedFunction(ex.core.BinaryOperator,S)
UnaryOperator = ex.WrappedFunction(ex.core.UnaryOperator,S)

from evaluator import WrappedRule,WrappedRuleEvaluator

Rule = WrappedRule(S)
RuleEvaluator = WrappedRuleEvaluator(S)

MatchCondition = ex.WrappedMatchCondition(S)
MulplicityList = ex.WrappedMulplicityList(S)



equal = BinaryOperator("=",ex.core.associative,ex.core.commutative,-6)


addition = BinaryOperator("+",ex.core.associative,ex.core.commutative,-11)
negative = UnaryOperator("-",ex.core.prefix,-12)
multiplication = BinaryOperator("*",ex.core.associative,ex.core.commutative,-13)
fraction = UnaryOperator("1/",ex.core.prefix,-14)
exponentiation = BinaryOperator("**",-15)

Group = ex.WrappedGroup(S)
Field = ex.WrappedField(S)


addition_group = Group(addition,negative,Zero)
multiplication_group = Group(multiplication,fraction,One)

real_field = Field(addition_group,multiplication_group)










def __latex_print_addition(printer,expr):
    neg_args = [arg for arg in expr.args if arg.function == negative]
    covered = set(neg_args)
    rest = [arg for  arg in expr.args if arg not in covered]
    rest_str = '+'.join(printer._printed_operator_arguments(expr,rest))
    if len(neg_args) == 0:
        return rest_str
    neg_str = '-'.join(printer._printed_operator_arguments(expr,[arg.args[0] for arg in neg_args]))
    return rest_str + '-' + neg_str

latex.register_printer(addition,__latex_print_addition)

def __latex_print_multiplication(printer,expr):
    denominators = [arg for arg in expr.args if arg.function == fraction]
    if len(denominators)>0:
        numerators = [arg for arg in expr.args if arg.is_atomic]
        if len(numerators) == 0:
            numerators = [One]
            
        covered = set(numerators + denominators)
        rest = [arg for  arg in expr.args if arg not in covered]

        denom_str = printer(multiplication(*[arg.args[0] for arg in denominators]))
        num_str =   printer(multiplication(*numerators))
        
        if len(rest) == 0:
            rest_str = ""
        elif len(rest) == 1:
            rest_str = printer(rest[0])
            if printer._needs_brackets_in(rest[0],expr):
                rest_str = printer._bracket_format() % rest_str
        else:
            rest = multiplication(*rest)
            rest_str =  printer(rest)
        
        return r'\frac{%s}{%s} \, %s ' % (num_str,denom_str,rest_str)

    is_numeric = lambda x: x.value != None or (x.function == exponentiation and x.args[0].value != None)
    
    numeric = [x for x in expr.args if is_numeric(x)]
    non_numeric = [x for x in expr.args if not is_numeric(x)]

    res  = '\cdot '.join(printer._printed_operator_arguments(expr,numeric)) 
    res += '\, '.join(printer._printed_operator_arguments(expr,non_numeric))

    return res

latex.register_printer(multiplication,__latex_print_multiplication)

def __latex_print_fraction(printer,expr):
    return r'\frac{1}{%s}' % printer(expr.args[0])

latex.register_printer(fraction,__latex_print_fraction)

# Here brackets are not set correctly
def __latex_print_exp(printer,expr):
    parg = printer._printed_operator_arguments(expr,end=-1)
    parg += [printer(expr.args[-1])]
    return '^'.join(['{%s}' % arg for arg in parg])

latex.register_printer(exponentiation,__latex_print_exp)



postorder_traversal = ex.wrapped_postorder_traversal(S)
preorder_traversal = ex.wrapped_postorder_traversal(S)





import expresso.visitor as visitor
from expresso.printer import Printer
from mpmath import mp

import functions as f
import expression as e

class CodePrinter(Printer):

    def __init__(self):
        super(CodePrinter,self).__init__(e.expression_converter)

    @visitor.on('expr',parent = Printer)
    def visit(self,expr):
        raise ValueError('cannot compile expression %s' % expr)
    
    def print_wildcard_symbol(self,expr):
        raise ValueError('cannot compile wildcard %s' % expr)

    def print_wildcard_function(self,expr):
        raise ValueError('cannot compile wildcard %s' % expr)

    @visitor.function(f.CustomFunction)
    def visit(self,expr):
        raise ValueError('cannot compile custom function %s' % expr)

    @visitor.function(f.ArrayAccess)
    def visit(self,expr):
        raise ValueError('cannot compile expression %s' % expr)

    @visitor.function(f.Tuple)
    def visit(self,expr):
        raise ValueError('cannot compile expression %s' % expr)





import expresso.pycas as pc
import rule_symbols as s

evaluator = pc.RewriteEvaluator(recursive=True,split_binary=True)

evaluator.add_rule(s.a*(s.b+s.c),s.a*s.b+s.a*s.c)
evaluator.add_rule(s.a**-1*(s.b+s.c)**-1,(s.a*s.b+s.a*s.c)**-1)

from .logic_evaluator import is_pure_numeric,is_atomic

is_positive_numeric_integer = pc.And(is_pure_numeric(s.n), pc.is_in_type(s.n,pc.Types.Integer), s.n > 1)

evaluator.add_rule((s.a+s.b)**s.n, (s.a+s.b) * (s.a+s.b)**(s.n-1), condition=is_positive_numeric_integer)
evaluator.add_rule((s.a+s.b)**-s.n, (s.a+s.b)**(-s.n+1)*(s.a+s.b)**-1,condition=is_positive_numeric_integer)

evaluator.add_rule(s.x*s.x,s.x**2,condition=is_atomic(s.x))
evaluator.add_rule(s.x*s.x**s.n,s.x**(s.n+1),condition=is_atomic(s.x))
evaluator.add_rule(s.x**s.m*s.x**s.n,s.x**(s.n+s.m),condition=is_atomic(s.x))

evaluator.add_rule((s.a*s.b)**s.c,s.a**s.c*s.b**s.c)

from .canonical_form import canonical_form
from .logic_evaluator import logic_evaluator
from .numeric_evaluator import numeric_evaluator
from .type_evaluator import type_evaluator
from .main_evaluator import evaluator as main_evaluator

expand_evaluator = pc.MultiEvaluator(recursive = True, split_binary=True)
expand_evaluator.add_evaluator(canonical_form)
expand_evaluator.add_evaluator(numeric_evaluator)
expand_evaluator.add_evaluator(evaluator)
expand_evaluator.add_evaluator(main_evaluator)
expand_evaluator.add_evaluator(type_evaluator)
expand_evaluator.add_evaluator(logic_evaluator)






def get_symbols_in(expr):
    import expresso.pycas as pc

    symbols = set()
    for e in pc.postorder_traversal(expr):
        if e.is_symbol:
            symbols.add(e)
    return symbols

def apply_all_arguments(expr,function):
    return expr.function(*[function(arg) for arg in expr.args])

def create_conditional_factor_evaluator(condition):

    from .expression import MulplicityList,MultiEvaluator,RewriteEvaluator,wildcard_symbols
    from .functions import multiplication_group,exponentiation,real_field

    from .evaluators.main_evaluator import evaluator as main_evaluator

    a,b,c,d = wildcard_symbols('a,b,c,d')

    def group_evaluator(m):
        e1 = m[a]
        e2 = m[b]
        m1 = MulplicityList(e1,multiplication_group,exponentiation,real_field)
        m2 = MulplicityList(e2,multiplication_group,exponentiation,real_field)
        common = m1.intersection(m2)

        commondict = dict(common)

        if len(common) == 0:
            return False
        else:
            if condition(commondict):
                m[a] = (m1 - common).as_expression()
                m[b] = (m2 - common).as_expression()
                m[c] = common.as_expression()
                return True
        return False

    evaluator = RewriteEvaluator(recursive=True,split_binary=True)
    evaluator.add_rule(a+b,c*(a+b),group_evaluator)
    evaluator.add_rule(a-b,c*(a-b),group_evaluator)
    evaluator.add_rule(-a+b,c*(-a+b),group_evaluator)
    evaluator.add_rule(-a-b,-c*(a+b),group_evaluator)

    from evaluators.canonical_form import canonical_form,format_evaluator
    from evaluators.logic_evaluator import logic_evaluator
    from evaluators.numeric_evaluator import numeric_evaluator
    from evaluators.type_evaluator import type_evaluator
    from evaluators.main_evaluator import evaluator as main_evaluator

    conditional_factor = MultiEvaluator(recursive = True, split_binary=True)
    conditional_factor.add_evaluator(canonical_form)
    conditional_factor.add_evaluator(numeric_evaluator)
    conditional_factor.add_evaluator(main_evaluator)
    conditional_factor.add_evaluator(evaluator)
    conditional_factor.add_evaluator(type_evaluator)
    conditional_factor.add_evaluator(logic_evaluator)

    def conditional_evaluate(expr,**kwargs):
        return format_evaluator(conditional_factor(expr,**kwargs),**kwargs)

    return conditional_evaluate

def expand(*args,**kwargs):
    from .evaluators import expand
    return expand(*args,**kwargs)

def split(args,condition):
    r1 = []
    r2 = []
    for arg in args:
        if condition(arg):
            r1.append(arg)
        else:
            r2.append(arg)
    return r1,r2

def term_decomposition(expr):
    import expresso.pycas as pc

    if expr.function == pc.multiplication:
        return [item for term in expr.args for item in term_decomposition(term)]
    elif expr.function == pc.exponentiation:
        e = expr.args[1]
        return [term**e for term in term_decomposition(expr.args[0])]
    elif expr.function == pc.negative:
        return [-1] + term_decomposition(expr.args[0])
    elif expr.function == pc.fraction:
        return [1/term for term in term_decomposition(expr.args[0])]
    else:
        return [expr]

def get_coefficient(expr,term):
    t = split(expr.args,lambda e:term in term_decomposition(e))[0]
    if len(t) == 0:
        return 0
    return (expr.function(*t)/term).evaluate()








import expresso.pycas as pc
import canonical_form
import logic_evaluator
import numeric_evaluator
import type_evaluator
import main_evaluator
import expand_evaluator

__cached_evaluators = {}
__use_global_cache = False

def evaluate(expr,context = pc.global_context,cache = None,format = True):

    if context in __cached_evaluators:
        main = __cached_evaluators[context]
    else:
        main = pc.MultiEvaluator(recursive=True,split_binary=True)
        main.add_evaluator(context)
        main.add_evaluator(main_evaluator.main_evaluator)
        __cached_evaluators[context] = main

    if cache == None and __use_global_cache:
        cache = __cached_evaluators

    expr = main(expr,cache = cache)
    if format:
        expr = canonical_form.format_evaluator(expr,cache = cache)
    return expr

def use_global_cache(v):
    __use_global_cache = v

def set_debug(v):

    def callback(r,m):
        from IPython.display import display_latex
        import sys

        print "appy rule: %s" % r

        lt = pc.latex(r.search.subs(m,evaluate=False)),\
             pc.latex(r.replacement.subs(m,evaluate=False)),\
             r"\;\text{ if }\;%s" % pc.latex(r.condition.subs(m,evaluate=False)) if r.condition is not None else ''

        display_latex(r"$$%s \rightarrow %s%s$$" % lt,raw=True)
        sys.stdout.flush()

    if v:
        main_evaluator.main_evaluator.set_rule_callback(callback)
        canonical_form.format_evaluator.set_rule_callback(callback)
        expand_evaluator.expand_evaluator.set_rule_callback(callback)
    else:
        main_evaluator.main_evaluator.set_rule_callback(None)
        canonical_form.format_evaluator.set_rule_callback(None)
        expand_evaluator.expand_evaluator.set_rule_callback(None)

def expand(expr,**kwargs):
    return canonical_form.format_evaluator(expand_evaluator.expand_evaluator(expr,**kwargs),**kwargs)





import expresso.visitor as visitor
from .. import expression as e
from .. import functions as f
from mpmath import mp

class LambdaCompiler(object):

    def __init__(self,value_converter = lambda x:x,function_module = None,value_module = None,cache=None,folding=False):
        self.dispatcher = self.visit.dispatcher
        self.value_converter = value_converter
        self.folding = folding
        if function_module == None:
            import math
            function_module = math
        if value_module == None:
            value_module = function_module
        self.function_module = function_module
        self.value_module = value_module

        if cache:
            self.cache = cache
        else:
            self.cache = {}

    def logic_reduce(self,op,args):
        return reduce(op,args)

    @visitor.on('expr')
    def visit(self,expr):
        raise ValueError('cannot compile expression: %s' % expr)

    def get_function(self,name):
        try:
            return getattr(self.function_module,name)
        except:
            return None

    @visitor.function
    def visit(self,expr):
        func = self.get_function(expr.function.name)
        if func == None:
            raise ValueError('unknown function: %s' % expr.function.name)
        cargs = [self.visit(arg) for arg in expr.args]
        return lambda args:func(*[arg(args) for arg in cargs])

    @visitor.function(f.OuterPiecewise)
    def visit(self,expr):
        return self.visit(expr.args[0])

    @visitor.function(f.addition)
    def visit(self,expr):
        cargs = [self.visit(arg) for arg in expr.args]
        return lambda args:sum([arg(args) for arg in cargs])

    @visitor.function(f.multiplication)
    def visit(self,expr):
        from operator import mul
        cargs = [self.visit(arg) for arg in expr.args]
        return lambda args:reduce(mul,[arg(args) for arg in cargs])

    @visitor.function(f.negative)
    def visit(self,expr):
        arg = self.visit(expr.args[0])
        return lambda args:-arg(args)

    @visitor.function(f.fraction)
    def visit(self,expr):
        arg = self.visit(expr.args[0])
        one = self.value_converter(1.)
        return lambda args:one/arg(args)

    @visitor.function(f.exponentiation)
    def visit(self,expr):
        from operator import pow
        cargs = [self.visit(arg) for arg in expr.args]
        return lambda args:reduce(pow,[arg(args) for arg in cargs])

    @visitor.function(f.equal)
    def visit(self,expr):
        from operator import eq
        cargs = [self.visit(arg) for arg in expr.args]
        return lambda args:self.logic_reduce(eq,[arg(args) for arg in cargs])

    @visitor.function(f.unequal)
    def visit(self,expr):
        cargs = [self.visit(arg) for arg in expr.args]
        return lambda args: self.logic_reduce(lambda x,y:x!=y ,[arg(args) for arg in cargs])

    @visitor.function(f.Less)
    def visit(self,expr):
        from operator import lt
        cargs = [self.visit(arg) for arg in expr.args]
        return lambda args:self.logic_reduce(lt,[arg(args) for arg in cargs])

    @visitor.function(f.Greater)
    def visit(self,expr):
        from operator import gt
        cargs = [self.visit(arg) for arg in expr.args]
        return lambda args:self.logic_reduce(gt,[arg(args) for arg in cargs])

    @visitor.function(f.LessEqual)
    def visit(self,expr):
        from operator import le
        cargs = [self.visit(arg) for arg in expr.args]
        return lambda args:self.logic_reduce(le,[arg(args) for arg in cargs])

    @visitor.function(f.GreaterEqual)
    def visit(self,expr):
        from operator import ge
        cargs = [self.visit(arg) for arg in expr.args]
        return lambda args:self.logic_reduce(ge,[arg(args) for arg in cargs])

    @visitor.function(f.And)
    def visit(self,expr):
        from operator import __and__
        cargs = [self.visit(arg) for arg in expr.args]
        return lambda args:self.logic_reduce(__and__,[arg(args) for arg in cargs])

    @visitor.function(f.Or)
    def visit(self,expr):
        from operator import __or__
        cargs = [self.visit(arg) for arg in expr.args]
        return lambda args:self.logic_reduce(__or__,[arg(args) for arg in cargs])

    @visitor.function(f.Xor)
    def visit(self,expr):
        from operator import __xor__
        cargs = [self.visit(arg) for arg in expr.args]
        return lambda args:self.logic_reduce(__xor__,[arg(args) for arg in cargs])

    @visitor.function(f.real)
    def visit(self,expr):
        arg = self.visit(expr.args[0])
        return lambda args:arg(args).real

    @visitor.function(f.imag)
    def visit(self,expr):
        arg = self.visit(expr.args[0])
        return lambda args:arg(args).imag

    @visitor.function(f.conjugate)
    def visit(self,expr):
        arg = self.visit(expr.args[0])
        return lambda args:arg(args).conjugate()

    @visitor.function(f.Abs)
    def visit(self,expr):
        arg = self.visit(expr.args[0])
        def evaluate(args):
            return abs(arg(args))
        return evaluate

    @visitor.function(f.Min)
    def visit(self,expr):
        arguments = [self.visit(arg) for arg in expr.args]
        return lambda args:min(*[arg(args) for arg in arguments])

    @visitor.function(f.Max)
    def visit(self,expr):
        arguments = [self.visit(arg) for arg in expr.args]
        return lambda args:max(*[arg(args) for arg in arguments])

    @visitor.function(f.Not)
    def visit(self,expr):
        arg = self.visit(expr.args[0])
        return lambda args:not arg(args)

    @visitor.function(f.unfoldable)
    def visit(self,expr):
        if self.folding:
            raise RuntimeError('folding unfoldable value')
        return self.visit(expr.args[0])

    @visitor.function(f.Tuple)
    def visit(self,expr):
        arguments = [self.visit(arg) for arg in expr.args]
        def make_tuple(args):
            return tuple((arg(args) for arg in arguments))
        return make_tuple

    @visitor.function(f.ArrayAccess)
    def visit(self,expr):

        array = expr.args[0].value
        indices = [self.visit(arg) for arg in expr.args[1:]]

        def access(args):
            idx = [int(arg(args)) for arg in indices[::-1]]
            for i,s in zip(idx,array.shape):
                if (i<0 or i>=s):
                    return self.value_converter(0)

            return array[tuple(idx)]

        return access

    @visitor.function(f.InnerPiecewise)
    def visit(self,expr):
        cond_args = [self.visit(arg.args[1]) for arg in expr.args]
        eval_args = [self.visit(arg.args[0]) for arg in expr.args]

        def evaluate(args):
            for i in xrange(len(cond_args)):
                if cond_args[i](args):
                    return eval_args[i](args)
            return 0

        return evaluate

    @visitor.symbol
    def visit(self,expr):
        name = expr.name

        def get_symbol(args):
            try:
                return self.value_converter(args[name])
            except KeyError:
                raise ValueError('undefined symbol %r' % name)

        return get_symbol

    @visitor.obj(f.SymbolicConstant)
    def visit(self,expr):
        try:
            v = self.value_converter(self.value_module.__dict__[expr.value.name])
        except KeyError:
            raise ValueError('cannot compile %s' % expr)
        return lambda args: v

    @visitor.atomic(e.S(True))
    def visit(self,expr):
        return lambda args:True

    @visitor.atomic(e.S(False))
    def visit(self,expr):
        return lambda args:False

    @visitor.atomic(e.I)
    def visit(self,expr):
        im = self.value_converter(1j)
        return lambda args:im

    @visitor.obj(e.Number)
    def visit(self,expr):
        v = self.value_converter(expr.value)
        return lambda args:v

    @visitor.function(f.CustomFunction)
    def visit(self,expr):
        cargs = [self.visit(arg) for arg in expr.args[1:]]
        f = expr.args[0].value
        if not hasattr(f,'python_function'):
                raise ValueError("custom function %s has no attribute 'python_function'" % f.name)
        callback = f.python_function
        return lambda args:callback(*[arg(args) for arg in cargs])

    @visitor.obj
    def visit(self,expr):
        value = self.value_converter(expr.value)
        return lambda args:value

def lambdify(expr,**kwargs):
    compiler = LambdaCompiler(**kwargs)
    compiled = compiler.visit(e.S(expr))
    return lambda **args:compiled(args)

def mpmathify(expr,**kwargs):
    from mpmath import mp
    vc = lambda x:mp.mpc(x) if isinstance(x,(complex,mp.mpc)) else mp.mpf(x)
    compiler = LambdaCompiler(value_converter=vc,function_module=mp,**kwargs)
    f = compiler.visit(e.S(expr))
    return lambda **args:f(args)

def N(expr,mp_dps = None,**kwargs):
    from mpmath import mp
    if mp_dps != None:
        mp.dps = mp_dps;
    f = mpmathify(expr,**kwargs)
    res = f()
    return res




from expression import *
import visitor

__all__ = ["Printer","LatexPrinter"]

class Printer(visitor.visitor_class()):

    def __init__(self,S):
        self.S = S

    def format_name(self,name):
        return name

    def register_printer(self,expr,printer):
        self.dispatcher.register_target(expr,printer)
    
    def needs_brackets_in(self,expr,parent):
        if expr.is_atomic:
            return False
        if not expr.function.is_operator and parent.function.is_operator:
            return False
        return parent.function.precedence <= expr.function.precedence
    
    def print_symbol(self,expr):
        return self.format_name(expr.name)
    
    def print_wildcard_symbol(self,expr):
        return self.print_symbol(expr)

    def print_wildcard_function(self,expr):
        return self.print_function(expr)
    
    def bracket_format(self):
        return "(%s)"
    
    def function_format(self):
        return "%s(%s)"

    def print_operator_argument(self,expr,parent):
        if self.needs_brackets_in(expr,parent):
            return self.bracket_format() % self(expr)
        return self(expr)
    
    def print_prefix_operator(self,expr,symbol):
        return symbol + self.print_operator_argument(expr.args[0],expr)

    def print_postfix_operator(self,expr,symbol):
        return self.print_operator_argument(expr.args[0],expr) + symbol
    
    def print_unary_operator(self,expr,symbol = None):
        f = expr.function
        if symbol == None:
            symbol = f.symbol
        if f.is_prefix:
            return self.print_prefix_operator(expr,symbol)
        else:
            return self.print_postfix_operator(expr,symbol)

    def printed_operator_arguments(self,parent,args=None,begin=0,end=None):
        if args == None:
            args = parent.args
        printed_args = [self.print_operator_argument(arg,parent) for arg in args[begin:end]]
        if parent.function.is_commutative:
            return sorted(printed_args)
        return printed_args

    def print_binary_operator(self,expr,symbol = None):
        if symbol == None:
            symbol = expr.function.symbol
        return symbol.join(self.printed_operator_arguments(expr))
    
    def print_function(self,expr,name = None):
        if name == None:
            name = self.format_name(expr.function.name)
        return self.function_format() % (name,','.join([ self(e) for e in expr.args ]))
        
    @visitor.on('expr')
    def visit(self,expr):
        raise ValueError('cannot print expression %s' % expr.name)

    @visitor.function
    def visit(self,expr):
        return self.print_function(expr)

    @visitor.atomic
    def visit(self,expr):
        return self.print_symbol(expr)

    @visitor.obj
    def visit(self,expr):
        return str(expr.value)

    @visitor.wildcard_symbol
    def visit(self,expr):
        return self.print_wildcard_symbol(expr)

    @visitor.wildcard_function
    def visit(self,expr):
        return self.print_wildcard_function(expr)

    @visitor.binary_operator
    def visit(self,expr):
        if len(expr.args) >= 2:
            return self.print_binary_operator(expr)
        else:
            return self.print_function(expr)

    @visitor.unary_operator
    def visit(self,expr):
        return self.print_unary_operator(expr)

    def __call__(self,expr):
        return self.dispatcher(self,self.S(expr))
    
class LatexPrinter(Printer):

    def __init__(self,*args,**kwargs):
        super(LatexPrinter,self).__init__(*args,**kwargs)

        self.latex_replacements = {}
        self.latex_replacements.update({w:r'\%s ' % w for w in ['Delta', 'Gamma', 'Lambda', 'Omega', 'Phi', 'Pi', 'Psi', 'Sigma', 'Theta', 'Upsilon', 'Xi', 'alpha', 'beta', 'chi', 'delta', 'epsilon', 'eta', 'gamma', 'kappa', 'lambda', 'mu', 'nabla', 'nu', 'omega', 'phi', 'pi', 'psi', 'rho', 'sigma', 'tau', 'theta', 'upsilon', 'varepsilon', 'varphi', 'varpi', 'varrho', 'varsigma', 'vartheta', 'xi', 'zeta', 'hbar']})

    @visitor.on('expr',parent = Printer)
    def visit(self,expr):
        raise ValueError('cannot print expression %s' % expr.name)
        
    def print_binary_operator(self,expr,symbol = None):
        if symbol == None:
            symbol = " " + expr.function.symbol + " "
        return symbol.join(self.printed_operator_arguments(expr))
        
    def bracket_format(self):
        return r"\left( %s \right) "
    
    def print_wildcard_symbol(self,expr):
        return '\mathbf{%s}' % self.format_name(expr.name[1:])

    def function_format(self):
        return r"%s \mathopen{} \left(%s \right) \mathclose{} "

    def format_name(self,name):

        underscore_parts = name.split('_')

        for i,p in enumerate(underscore_parts):
            space_parts = p.split(' ')
            for i2,p2 in enumerate(space_parts):
                if p2 in self.latex_replacements:
                    space_parts[i2] = self.latex_replacements[p2]
                elif len(p2) > 1:
                    space_parts[i2] = r'\text{%s} ' % p2
            underscore_parts[i] = '\; '.join(space_parts)

        return reduce(lambda p1,p2:'{%s}_{%s} ' % (p2,p1),underscore_parts[::-1])

    def print_wildcard_function(self,expr):
        f = expr.function
        name = self.format_name(f.name[1:])
        name = '\mathbf{%s} ' % name
        return self.print_function(expr,name=name)

    @visitor.obj
    def visit(self,expr):
        return self.format_name(expr.name)



from setuptools import setup, Extension, find_packages
from glob import glob

setup(
    name='expresso',
    version='0.1.8',
    description='A symbolic expression manipulation library.',
    license='MIT',
    author='Lars Melchior',
    author_email='thelartians@gmail.com',

    url='https://github.com/TheLartians/Expresso',
    
    packages=find_packages(exclude=['tests*']),
    
    keywords='expression symbolic manipulation algebra',

    extras_require={
        'pycas':['numpy','mpmath']
    },

    zip_safe=False,

    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7'
    ],

    ext_modules=[
        Extension('_expresso',
                  sources = glob('source/expresso/*.cpp') + ['libs/sha256/sha256.cpp','source/python.cpp'],
                  include_dirs=['source','source/expresso','libs'], 
                  libraries=['boost_python'], 
                  extra_compile_args=['-g','-std=c++11','-Wno-unknown-pragmas','-Wno-unused-local-typedef','-O3'] 
                  ),
        ],

    test_suite='nose.collector',
    tests_require=['nose'],

)



from unittest import TestCase

from expresso.pycas import *
import numpy as np

i,j = symbols('i,j',type=Types.Integer)
x,y = symbols("x, y",type=Types.Real)
z = Symbol('z',type=Types.Complex)

f = Function('f',argc = 1)
g = Function('g',argc = 2)

class TestAlgebra(TestCase):

    def test_factor(self):
        expr = ((-2*x**(z)*(y+2*y/6)/x**(z-1)-y*(x+x/3)+42*3*x**2/(3*x*y**((x-2*x)/x)))+4*x*y)/(x*y)
        self.assertEqual(expr.evaluate(), S(42))

    def test_derivative(self):
        expr = derivative(derivative(log(exp(x**y)*x**2),x),y)
        self.assertEqual(expr.evaluate(),x**(y-1)*(y*log(x)+1))


class TestCompilers(TestCase):

    def assertNumpyEqual(self,a,b):
        return self.assertSequenceEqual(list(a),list(b))

    def test_numpy_compiler(self):
        abs_expr = numpyfy(piecewise((x,x>=0),(-x,True)))
        self.assertNumpyEqual(abs_expr(x=[-np.pi,3,-2.5,0,np.e,np.pi,-42]),[np.pi,3,2.5,0,np.e,np.pi,42])

        random_numbers = np.random.rand(100)
        array_expr = numpyfy(array('rand',random_numbers)( floor(array('rand',random_numbers)(i)*100)),restype=float)

        self.assertNumpyEqual( array_expr(i = range(10)), random_numbers[(random_numbers[range(10)]*100).astype(int)] )

    def test_compiler_consistency(self):

        def my_function_python(x,y):
            return (abs(x)/(1+abs(y)))%1

        my_function_ccode = '''
        double myfunction(double x,double y){
            return fmod(abs(x)/(1+abs(y)),1);
        }
        '''

        pycas_custom_function = custom_function("myfunction",
                                        argc = 2,
                                        python_function=my_function_python,
                                        ccode = my_function_ccode,
                                        return_type=Types.Real)

        npx,npy = np.meshgrid(np.linspace(-10, 10, 1001),np.linspace(-9.5, 9.5, 950))

        snpx = array('np_x',npx+2*np.random.rand(*npy.shape))
        p = parameter('p',1)

        expr = piecewise((pi*snpx(sin(x)*cos(y)*1000,y*50),y>0),(e*pycas_custom_function(y-p*x,x+p*y)*10,True))

        fs_def = FunctionDefinition('f_single',(x,y),expr,return_type=Types.Real,parallel=False)
        fp_def = FunctionDefinition('f_parallel',(x,y),expr,return_type=Types.Real,parallel=True)

        clib = ccompile(fp_def,fs_def)
        nlib = ncompile(fp_def,fs_def)

        p.set_value(2)

        lf = lambdify(expr)
        N = mpmathify(expr)

        for (vx,vy) in zip(10*(np.random.rand(1000)-0.5),10*(np.random.rand(1000)-0.5)):
            self.assertTrue( np.isclose(nlib.f_single(vx,vy),lf(x=vx,y=vy)) )
            self.assertTrue( np.isclose(lf(x=vx,y=vy),float(N(x=vx,y=vy))) )

        lib_f_single_t = clib.f_single(npx,npy)
        lib_f_parallel_t = clib.f_parallel(npx,npy)
        cf_single_t = nlib.f_single(npx,npy)
        cf_parallel_t = nlib.f_parallel(npx,npy)

        self.assertTrue( np.allclose(lib_f_single_t, lib_f_parallel_t) )
        self.assertTrue( np.allclose(cf_single_t, cf_parallel_t) )
        self.assertTrue( np.allclose(cf_parallel_t, lib_f_single_t) )




import printer

from .expression import *
from .functions import *
from .compilers import *
from .algorithms import *


def use_global_cache(v):
    from .evaluators import use_global_cache
    use_global_cache(v)







import expresso.pycas as pc
import rule_symbols as s
from .logic_evaluator import is_numeric,is_atomic

from .canonical_form import format_evaluator
from .logic_evaluator import evaluator as logic_evaluator

from .logic_evaluator import is_mpmath

compile_evaluator = pc.RewriteEvaluator(recursive=False,split_binary=True)

fold_accuracy = 20

def fold(m):
    try:
        res = m[s.x].N(fold_accuracy,folding=True)
        if type(res) == tuple:
            return False
        m[s.y] = pc.expresso.create_object(res)
    except Exception as e:
        return False

compile_evaluator.add_rule(s.x,s.y,fold)
compile_evaluator.add_rule(s.x**2,s.x*s.x,condition=is_atomic(s.x))

compile_evaluator.add_rule(s.x*abs(s.y),abs(s.x*s.y),condition=is_mpmath(s.x))
compile_evaluator.add_rule(s.x*(s.y+s.z),(s.x*s.y+s.x*s.z),condition=is_mpmath(s.x))

compiler_opt_evaluator = pc.MultiEvaluator(recursive = True, split_binary=True)
compiler_opt_evaluator.add_evaluator(compile_evaluator)
compiler_opt_evaluator.add_evaluator(logic_evaluator)

def optimize_for_compilation(expr,cache = None,prec=20):
    global fold_accuracy
    fold_accuracy = prec
    opt = None

    # TODO: why do we need to evaluate multiple times?
    N = 10
    while opt != expr and N>0:
        N -= 1
        opt = expr
        expr = compiler_opt_evaluator(expr, cache = cache)

    return format_evaluator(expr, cache = cache)





from lambda_compiler import lambdify,mpmathify,N
from numpy_compiler import numpyfy,ncompile
from c_compiler import ccompile,c_complex


class FunctionDefinition(object):

    def __init__(self, name, args, expr, return_type = None, arg_types = None, parallel=True):
        from ..expression import S
        self.name = name
        self.expr = S(expr)
        self.args = args if isinstance(args,(list,tuple)) else (args,)
        self.return_type = return_type
        self.arg_types = arg_types
        self.parallel = parallel

    def __str__(self):
        return '%s(%s) = %s' % (self.name,','.join([str(arg) for arg in self.args]),self.expr)






import expresso.pycas as pc
import rule_symbols as s

ordered_types = (pc.Types.Boolean,pc.Types.Natural,pc.Types.Integer,pc.Types.Rational,pc.Types.Real,pc.Types.Complex)

evaluator = pc.RewriteEvaluator(recursive=True, split_binary=True)

from .logic_evaluator import is_explicit_natural,is_function_type

evaluator.add_rule(pc.DominantType(s.x),s.x)

for i in range(len(ordered_types)):
    evaluator.add_rule(pc.DominantType(pc.Types.Imaginary,ordered_types[i]),pc.Types.Complex)
    evaluator.add_rule(pc.Type(ordered_types[i]),pc.Types.Type)
    for j in range(i):
      evaluator.add_rule(pc.DominantType(ordered_types[j],ordered_types[i]),ordered_types[i])

def eval_type_equality(m):
    tx = m[s.x]
    ty = m[s.y]
    if tx in ordered_types and ty in ordered_types:
        m[s.z] = tx == ty
        return True
    return False

evaluator.add_rule(pc.equal(s.x,s.y),False,eval_type_equality)

evaluator.add_rule(pc.Type(pc.Types.Imaginary*pc.Types.Complex),pc.Types.Complex)
evaluator.add_rule(pc.DominantType(s.x,s.x),s.x)

evaluator.add_rule(pc.Type(pc.Types.Imaginary),pc.Types.Type)
evaluator.add_rule(pc.Type(pc.Type(s.x)),pc.Types.Type)

evaluator.add_rule(pc.Type(True),pc.Types.Boolean)
evaluator.add_rule(pc.Type(False),pc.Types.Boolean)
evaluator.add_rule(pc.Type(s.x),pc.Types.Natural,condition=is_explicit_natural(s.x))
evaluator.add_rule(pc.Type(1/s.x),pc.DominantType(pc.Types.Rational,pc.Type(s.x)))

evaluator.add_rule(pc.Type(s.x**s.y),pc.OperationType(pc.Type(s.x)**pc.Type(s.y)))
evaluator.add_rule(pc.Type(s.x*s.y),pc.OperationType(pc.Type(s.x)*pc.Type(s.y)))

for t in (pc.Types.Natural,pc.Types.Integer,pc.Types.Rational,pc.Types.Real):
    evaluator.add_rule(pc.OperationType(pc.Types.Imaginary*t),pc.Types.Imaginary)
evaluator.add_rule(pc.OperationType(pc.Types.Imaginary*pc.Types.Imaginary),pc.Types.Real)

evaluator.add_rule(pc.OperationType(s.x**pc.Types.Natural),s.x)
evaluator.add_rule(pc.OperationType(s.x**pc.Types.Integer),pc.DominantType(s.x,pc.Types.Rational))
evaluator.add_rule(pc.OperationType(pc.Types.Natural**pc.Types.Rational),pc.Types.Real)

evaluator.add_rule(pc.OperationType(s.x**s.y),pc.Types.Complex)


evaluator.add_rule(pc.OperationType(s.x*s.y),pc.DominantType(s.x,s.y),condition=pc.Not(pc.Or(is_function_type(s.x,pc.Type),is_function_type(s.y,pc.Type))))
evaluator.add_rule(pc.OperationType(s.x)**s.y,pc.DominantType(s.x,s.y),condition=pc.Not(pc.Or(is_function_type(s.x,pc.Type),is_function_type(s.y,pc.Type))))


evaluator.add_rule(pc.Type(s.x+s.y),pc.DominantType(pc.Type(s.x),pc.Type(s.y)))
evaluator.add_rule(pc.Type(-s.x),pc.DominantType(pc.Types.Integer,pc.Type(s.x)))
evaluator.add_rule(pc.Type(pc.pi),pc.Types.Real)
evaluator.add_rule(pc.Type(pc.e),pc.Types.Real)
evaluator.add_rule(pc.Type(pc.I),pc.Types.Imaginary)

evaluator.add_rule(pc.Type(pc.factorial(s.x)),pc.Types.Natural)
evaluator.add_rule(pc.Type(pc.sign(s.x)),pc.Types.Integer)
evaluator.add_rule(pc.Type(pc.floor(s.x)),pc.Types.Integer)
evaluator.add_rule(pc.Type(pc.ceil(s.x)),pc.Types.Integer)
evaluator.add_rule(pc.Type(pc.round(s.x)),pc.Types.Integer)
evaluator.add_rule(pc.Type(pc.mod(s.x,s.y)),pc.Types.Integer)

evaluator.add_rule(pc.Type(pc.Abs(s.x)),pc.OperationType(pc.Abs(pc.Type(s.x))))
evaluator.add_rule(pc.OperationType(pc.Abs(pc.Types.Complex)),pc.Types.Real)
evaluator.add_rule(pc.OperationType(pc.Abs(pc.Types.Imaginary)),pc.Types.Real)
evaluator.add_rule(pc.OperationType(pc.Abs(pc.Types.Real)),pc.Types.Real)
evaluator.add_rule(pc.OperationType(pc.Abs(pc.Types.Rational)),pc.Types.Rational)
evaluator.add_rule(pc.OperationType(pc.Abs(pc.Types.Integer)),pc.Types.Natural)
evaluator.add_rule(pc.OperationType(pc.Abs(pc.Types.Natural)),pc.Types.Natural)


evaluator.add_rule(pc.Type(pc.real(s.x)),pc.Types.Real)
evaluator.add_rule(pc.Type(pc.imag(s.x)),pc.Types.Real)
evaluator.add_rule(pc.Type(pc.conjugate(s.x)),pc.Types.Complex)

evaluator.add_rule(pc.Type(pc.Indicator(s.x)),pc.Types.Natural)
evaluator.add_rule(pc.Type(pc.OuterPiecewise(s.x)),pc.Type(s.x))
evaluator.add_rule(pc.Type(pc.InnerPiecewise((s.a,s.b),s.x)),pc.DominantType(pc.Type(s.a),pc.Type(pc.InnerPiecewise(s.x))))
evaluator.add_rule(pc.Type(pc.InnerPiecewise((s.a,s.b))),pc.Type(s.a))


evaluator.add_rule(pc.Type(pc.derivative(s.x,s.y)),pc.Type(s.x))
evaluator.add_rule(pc.Type(pc.evaluated_at(s.x,s.y,s.z)),pc.DominantType(pc.Type(s.x),pc.Type(s.z)))

evaluator.add_rule(pc.Type(pc.tmp(s.x)),pc.Type(s.x))
evaluator.add_rule(pc.Type(pc.sqrt(s.x)),pc.Type(s.x**(1/pc.S(2))))

evaluator.add_rule(pc.Type(pc.atan2(s.x,s.y)),pc.DominantType(pc.Type(s.x),pc.Type(s.x),pc.Types.Rational))

for f in [pc.exp,pc.log,pc.sin,pc.cos,pc.asin ,pc.acos,pc.tan,pc.atan,pc.cot,pc.acot,pc.sinh,pc.cosh,pc.asinh,pc.acosh,pc.tanh,pc.atanh,pc.coth,pc.acoth]:
    evaluator.add_rule(pc.Type(f(s.x)),pc.DominantType(pc.Type(s.x),pc.Types.Rational))

def issubtype(x,t):
    return pc.equal(pc.DominantType(pc.Type(x),t),t)

from .logic_evaluator import is_mpmath
from mpmath import mp

def mpmath_type_evaluator(m):
    v = m[s.x].value
    if isinstance(v,mp.mpf):
        m[s.y] = pc.Types.Real
    elif isinstance(v,mp.mpc):
        m[s.y] = pc.Types.Complex
    else:
        raise AttributeError('unknown mpmath type')

evaluator.add_rule(pc.Type(s.x),s.y,condition=is_mpmath(s.x))


from .canonical_form import canonical_form
from .logic_evaluator import logic_evaluator

type_evaluator = pc.MultiEvaluator(recursive=True, split_binary=True)
type_evaluator.add_evaluator(canonical_form)
type_evaluator.add_evaluator(logic_evaluator)
type_evaluator.add_evaluator(evaluator)




















